$input_file = shift;
$starting_tree = shift;
$outtree = shift;
$fs = shift;

#!/usr/bin/perl 
#
# Usman Roshan
# September 2004
# 
# derived from
#
# perlRat.pl v1.0
# Modified February 13, 2002 10:00
#  Olaf R.P. Bininda-Emonds

$filename = $input_file;
$NumRepl = 2;
$file_suffix = $fs;
$PercIncl = 0.25;
$weight = 2;
$NumChar = 0;
$verbose = 0;
$logfile = "ratchet.log";
$paupfile = "ratchet.nex";
$treefile = "ratchet.tre";
$contreefile = "conratchet.tre";

if ($file_suffix eq "") {
  print "No file suffix!\n";
  exit (1);
}
else {
  $logfile .= $file_suffix;
  $treefile .= $file_suffix;
  $contreefile .= $file_suffix;
  $paupfile .= $file_suffix;
}

open(IN, $filename); 
$l = <IN>;
my @temp = split(/\s+/, $l);
if ($temp[0] !~ /\d/) {
  $number_of_leaves = $temp[1];
}
else {
  $number_of_leaves = $temp[0];
}
my $taxa = "", $count = 0;
while($count < $number_of_leaves) {
  $_ = <IN>;
  @temp = split(/\s+/, $_);
  $taxa .= $temp[0]." ";
  $count++;
}
close IN;

if (not $NumChar and not $filename) {
  die "Must specify total number of characters or filename\n" 
}

# Get number of characters from file (if specified)
if ($filename) {
  open (fDATA, "<$filename") or die "Cannot open data file $filename\n";
  open (fCMD, ">$paupfile") or die "Cannot open batch file $paupfile \n";
  while (<fDATA>) {
    #chomp;
    #	print "$_\n";
    if ($_ =~ /\s*(\d+)\s+(\d+)/) {  #specific for Phylip file
      $NumChar = $2;
      print fCMD "#NEXUS\n";
      print fCMD "begin characters;\n";
      print fCMD "\tdimensions newtaxa ntax=$1 nchar=$2;\n";
      print fCMD "\tformat datatype=nucleotide gap=- missing=?;\n";
      print fCMD "\tmatrix\n";
    }
    else {
      print fCMD "$_\n";
    }
  }
  print fCMD ";\n";
  print fCMD "end;\n\n";
}

die "Error reading number of characters" if (not $NumChar);

open(START_TREE, $starting_tree); $st=<START_TREE>; close START_TREE;
if($st !~ /\(/) { print "no starting tree\n"; exit (1); }

print fCMD "begin tree;\n";
print fCMD "tree tree_0"." = \[&U\] $st;\n";
print fCMD "end;\n";

$NumIncl = int($NumChar * $PercIncl);
$PAUPseed=int(10000*rand());

print fCMD "[Ratchet parameters: NumChar = $NumChar\n";
print fCMD "                     Inclusion Percentage = $PercIncl ($NumIncl characters)\n";
print fCMD "                     Replicates = $NumRepl] \n\n";
print fCMD "begin paup;\n";
print fCMD "\tset autoclose=yes warntree=no warnreset=no notifybeep=no monitor=no;\n";
print fCMD "\tlog file=$logfile replace;\n";
print fCMD "\tset criterion=parsimony;\n";

print fCMD "\n\t[!][!*** Replicate 0 (initial tree) ***]\n\n";
print fCMD "\thsearch start=current swap=TBR multrees=no dstatus=60;\n";
print fCMD "\tsavetrees file=$treefile format=nexus replace;\n\n";

for (1..$NumChar) {		#Create array containing numbers of all characters
  push @MasterCharString, $_;
}

for (1..$NumRepl) {
  @InclList = (); 
  @CharString = @MasterCharString;
  
  for (1..$NumIncl)	{
    #Randomly select required proportion of characters from array without replacement
    push @InclList, splice(@CharString,rand @CharString,1);
  }
  
  @InclList = sort {$a <=> $b} @InclList;
  
  print fCMD "\t[!][!*** Replicate #$_ ***]\n\n";
  print fCMD "\tweights $weight: @InclList;\n";
  print fCMD "\thsearch start=current swap=TBR multrees=no dstatus=none;\n";
  print fCMD "\tweights 1:all;\n";
  print fCMD "\thsearch start=current swap=TBR multrees=no dstatus=60;\n";
  print fCMD "\tsavetrees file=$treefile format=nexus append;\n\n";
}
print fCMD "\t[!][!*** Determining strict consensus tree ***]\n\n";
print fCMD "\tgettrees file=$treefile allblocks=yes warntree=no;\n";
print fCMD "\t\tfilter best=yes;\n";
print fCMD "\t\tsavetrees file=$contreefile format=phylip brlens=yes replace;\n";
print fCMD "\n\tlog stop;\n";
print fCMD "end;\n\n";
print fCMD "quit warntsave=no;\n";
close fCMD;
 
system "paup -u -n $paupfile > junk$fs";
system "tail -n 1 $contreefile > $outtree";
system "/bin/rm $paupfile $logfile $contreefile $treefile junk$fs";
