#This file calls paup to compute the MP score of a tree. 
#
#Copyright (C) 2004 The University of Texas at Austin
#
#This program is free software; you can redistribute it and/or modify it
#under the terms of the GNU General Public License as published by the Free
#Software Foundation; either version 2 of the License, or (at your option)
#any later version.
#
#This program is distributed in the hope that it will be useful, but WITHOUT
#ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#more details.
#
#You should have received a copy of the GNU General Public License along with
#this program; if not, write to the Free Software Foundation, Inc., 59 Temple
#Place - Suite 330, Boston, MA 02111-1307, USA.
#
#Email: usman@cs.njit.edu
#Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
#Science Department, GITC 4400, University Heights, Newark, NJ 07102 

$seqfile = shift;
$treefile = shift;
$scorefile = shift;
$file_suffix = shift;

open(SEQ_FILE, $seqfile);
$l = <SEQ_FILE>;
@temp = split(/\s+/, $l);
if ($temp[0] !~ /\d/) {
  $number_of_leaves = $temp[1];
  $seq_length = $temp[2];
}
else {
  $number_of_leaves = $temp[0];
  $seq_length = $temp[1];
}

$taxa = "";
while(<SEQ_FILE>) {
  @temp = split(/\s+/, $_);
  $taxa .= $temp[0]." ";
}
close SEQ_FILE;

open(PAUP_FILE, ">paup_mpscore_input$file_suffix");
print PAUP_FILE "#NEXUS\n";
print PAUP_FILE "begin taxa;\ndimensions ntax=$number_of_leaves".";\n";
print PAUP_FILE "taxlabels $taxa".";\nend;\n";
print PAUP_FILE "begin characters;\ndimensions nchar=$seq_length".";\n";
print PAUP_FILE "format datatype=nucleotide gap=- missing=?;\nmatrix\n";

open(SEQ_FILE, $seqfile);
<SEQ_FILE>;

while (<SEQ_FILE>) {
  $line1 = $_;
  $line1 =~ s/\s+/ /g;
  @temp = split(/\s/, $line1);
  $taxa = $temp[0];
  $seq = $temp[1];
  $seq = substr($seq, 0, $seq_length);
  print PAUP_FILE $taxa . " " . $seq . "\n";
}
print PAUP_FILE ";\nend;\n";
close SEQFILE;

srand;
$x = rand 100000;
$seed = int $x;

print PAUP_FILE "begin tree;\n";

open (MODEL_TREE, $treefile);
my $tree = <MODEL_TREE>;
close MODEL_TREE;

print PAUP_FILE "tree tree_0"." = \[&U\] $tree;\n";
print PAUP_FILE "end;\n";

print PAUP_FILE "begin paup;\n";
print PAUP_FILE "set monitor=no;\n";
print PAUP_FILE "pset collapse=no;\n";
print PAUP_FILE "pscores all/scorefile=$scorefile replace=yes;\n";
print PAUP_FILE "savetrees format=phylip brlens=yes file=$treefile replace;\n";
print PAUP_FILE "quit;\nend;\n";
close PAUP_FILE;

system "paup paup_mpscore_input$file_suffix -n";
system "/bin/rm paup_mpscore_input$file_suffix";
