#include<ugraph.h>
#include<stdlib.h>
#include<stdio.h>
#include<assert.h>

node new_node(ugraph G, void* aux_vars_) {
  
  node n;

  n = (node) malloc(sizeof(struct node_struct));
  
  if(!G->first_node) G->first_node = n;
  if(G->last_node) G->last_node->next = n;

  n->prev = G->last_node;
  n->next = 0;
  G->last_node = n;

  G->n_nodes++;
  G->id_nodes++;
  n->id = G->id_nodes;
  n->degree = 0;
  
  n->first_edge = 0;
  n->last_edge = 0;
  
  n->aux_vars = aux_vars_;
  
  return n;
}

void del_edge(ugraph G, edge e){

  e->u->degree--;
  e->v->degree--;

  if(e == G->first_edge) G->first_edge = e->next;
  if(e == G->last_edge) G->last_edge = e->prev;

  if(e == e->u->first_edge) e->u->first_edge = e->u_next;
  if(e == e->u->last_edge) e->u->last_edge = e->u_prev;

  if(e == e->v->first_edge) e->v->first_edge = e->v_next;
  if(e == e->v->last_edge) e->v->last_edge = e->v_prev;
  
  if(e->prev) e->prev->next = e->next;
  if(e->next) e->next->prev = e->prev;

  if(e->u_prev) {
    if(e->u_prev->u == e->u) e->u_prev->u_next = e->u_next;
    else if(e->u_prev->v == e->u) e->u_prev->v_next = e->u_next;
    else assert(0);
  }
  if(e->v_prev) {
    if(e->v_prev->v == e->v) e->v_prev->v_next = e->v_next;
    else if(e->v_prev->u == e->v) e->v_prev->u_next = e->v_next;
    else assert(0);
  }
  
  if(e->u_next) {
    if(e->u_next->u == e->u) e->u_next->u_prev = e->u_prev;
    else if(e->u_next->v == e->u) e->u_next->v_prev = e->u_prev;
    else assert(0);
  }
  
  if(e->v_next){
    if(e->v_next->v == e->v) e->v_next->v_prev = e->v_prev;
    else if(e->v_next->u == e->v) e->v_next->u_prev = e->v_prev;
    else assert(0);
  }
  
  if(e->aux_vars) free(e->aux_vars);
  free(e);
  G->n_edges--;
}
  
void del_node(ugraph G, node u){ 

  edge temp_edge = u->first_edge;
  while(temp_edge){
    del_edge(G, temp_edge);
    temp_edge = u->first_edge;
  }
  
  if(u->prev) u->prev->next = u->next;
  if(u->next) u->next->prev = u->prev;

  if(G->first_node == u) G->first_node = u->next;
  if(G->last_node == u) G->last_node = u->prev;

  if(u->aux_vars) free(u->aux_vars);
  free(u);
  G->n_nodes--;
}

edge new_edge(ugraph G, node u, node v, void* aux_vars_) {
  
  edge e;
  
  e = (edge) malloc(sizeof(struct edge_struct));

  e->aux_vars = aux_vars_;

  if(!G->first_edge) G->first_edge = e;
  if(G->last_edge) G->last_edge->next = e;

  e->prev = G->last_edge;
  e->next = 0;
  G->last_edge = e;

  G->n_edges++;
  G->id_edges++;
  e->id = G->id_edges;

  u->degree++;
  v->degree++;

  e->u = u;
  e->v = v;
  
  if(u->last_edge) {
    e->u_next = 0;
    if(u->last_edge->u == u)
      u->last_edge->u_next = e;
    else if(u->last_edge->v == u)
      u->last_edge->v_next = e;
    else assert(0);
    
    e->u_prev = u->last_edge; 
    u->last_edge = e;
  }
  else {
    e->u_next = 0; e->u_prev = 0;
    u->first_edge = e;
    u->last_edge = e;
  }
  
  if(v->last_edge) {
    e->v_next = 0;

    if(v->last_edge->v == v)
      v->last_edge->v_next = e;
    else if(v->last_edge->u == v)
      v->last_edge->u_next = e;
    else assert(0);

    e->v_prev = v->last_edge; 
    v->last_edge = e;
  }
  else {
    e->v_next = 0; e->v_prev = 0;
    v->last_edge = e;
    v->first_edge = e;
  }

  return e;
}

ugraph new_graph() {
  
  ugraph G;
   
  G = (ugraph) malloc(sizeof(struct ugraph_struct));
  
  G->first_edge = 0;
  G->last_edge = 0;
  G->first_node = 0;
  G->last_node = 0;
  
  G->n_nodes = 0;
  G->n_edges = 0;
  
  G->id_edges = 0;
  G->id_nodes = 0;

  return G;
}

void del_graph(ugraph G) {
  
  while(G->first_node) del_node(G, G->first_node);
  /*while(G->first_edge) del_edge(G->first_edge);*/
}

void print_graph(ugraph G) {
  
  /* Data */
  node n, n2; edge e;

  /* Code */
  printf("\nEdges\n");
  for(e = G->first_edge; e; e = e->next){
    printf("ptr: %d id: %d u_id: %d v_id: %d\nprev: %d next: %d\nprev_u: %d next_u: %d\nprev_v: %d next_v: %d\n\n",
	   (long)e, e->id, e->u->id, e->v->id, (long)e->prev, (long)e->next, 
	   (long)e->u_prev, (long)e->u_next, (long)e->v_prev, (long)e->v_next);
  }

  printf("Num of nodes = %d, num of edges = %d\n", G->n_nodes, G->n_edges);
  for(n = G->first_node; n; n = n->next) {
    printf("(%d %d): ", n->id, n->degree);
    for(e = n->first_edge; e;){
      printf(" (%d-%d) ", e->u->id, e->v->id);
      if(e->u == n) e = e->u_next;
      else if (e->v == n) e = e->v_next;
      else assert(0);
    }
    printf("\n");
  }
}

node opposite(node u, edge e){ 
  
  assert(e->u == u || e->v == u);
  if(e->u == u) return e->v;
  else return e->u;
}
