#include<hashtable.h>
#include<string.h>
#include<stdlib.h>
#include<stdio.h>

int probelength_hashtable(hashtable ht){return ht->probelength_;}

void del_hashtable(hashtable ht){
  int i;
  for(i=0;i<ht->size_;i++)
    if(ht->table_[i]) free(ht->table_[i]);
  free(ht->table_);
}    

hashtable new_hashtable(long sz){
  int i;
  hashtable ht;
  
  ht = (hashtable) malloc(sizeof(struct hashtable_struct));
  ht->size_ = sz;

  /* allocate memory for table*/
  ht->table_ = (char**) malloc(ht->size_*sizeof(char*)); 

  /* initialize to NULLS*/
  for(i = 0; i < ht->size_; i++) ht->table_[i] = 0; 
  ht->P_ = 2147483647; /* initialize P (large prime)*/
  ht->probelength_ = 0;
  
  /*initialize hash1 and hash2 variables*/
  ht->a1_ = drand48() * (ht->P_ - 1) + 1;
  ht->a2_ = drand48() * (ht->P_ - 1) + 1;
  ht->b1_ = drand48() * ht->P_;
  ht->b2_ = drand48() * ht->P_;

  return ht;
}

/* This is called before hash1 and hash2. It sets the 
   value of h_ to be used by hash1 and hash2 and saves excessive calls
   to the hash function.*/
unsigned long hash(hashtable ht, const char* key) {
  unsigned long g;

  ht->h_ = 0;
  while(*key) {
    ht->h_ = (ht->h_ << 4) + *key++;
    g = ht->h_ & 0xF0000000L;
    if(g) ht->h_ ^= g >> 24;
    ht->h_ &= ~g;
  }
  return(ht->h_ % ht->P_); /*P_ is a large prime*/
}

unsigned long hash1(hashtable ht) {
  return ((((ht->a1_ * ht->h_) + ht->b1_) % ht->P_) % ht->size_) ;
}

unsigned long hash2(hashtable ht) {
  return ((((ht->a2_ * ht->h_ + ht->b2_) % ht->P_) % (ht->size_ - 3)) | 1);
}

void insert_hashtable(hashtable ht, char* key) {
  unsigned long i;
  int h1, h2;
  
  hash(ht, key);
  h1 = hash1(ht);
  h2 = hash2(ht);
  
  ht->probelength_ = 0;
  
  for(i = h1; ht->table_[i]; i = (i + h2) % ht->size_){
    ht->probelength_++;
    if(!strcmp(ht->table_[i],key)) return;
  }
  
  ht->table_[i] = key;

  return;
}
    
char* lookup_hashtable(hashtable ht, char* key) {
  unsigned long i;
  int h1, h2;
  
  hash(ht, key);
  h1 = hash1(ht);
  h2 = hash2(ht);
  
  ht->probelength_ = 0;

  for(i = h1; ht->table_[i]; i = (i + h2) % ht->size_) {
    ht->probelength_++;
    if(!(ht->table_[i])) return 0;
    else if(ht->table_[i] && !strcmp(ht->table_[i],key)) 
      return ht->table_[i];
  }

  return 0;
}

