/*! \file hashtable.h
\brief Contains definition for a hash table structure which
uses strings as keys
*/

#ifndef HASHTABLE
#define HASHTABLE

typedef struct hashtable_struct* hashtable;

struct hashtable_struct { 
  
  /*size is size of table*/
  long size_;
  
  /*number of collisions before empty slot is found*/
  int probelength_;
  
  /* hash sets this value.  it is used in hash1 and hash2
     to avoid calling hash again (efficiency reasons) */
  unsigned long h_;
  
  /*random values used in hash functions.*/
  unsigned long a1_, a2_, b1_, b2_, P_;
  
  char** table_;
};

/*hash functions*/
unsigned long hash(hashtable ht, const char* str);
inline unsigned long hash1 (hashtable ht);
inline unsigned long hash2 (hashtable ht);

/*dictionary operations*/
void  insert_hashtable (hashtable ht, char* key);
char* lookup_hashtable (hashtable ht, char* key);

int probelength_hashtable(hashtable ht);

/*constructor*/
hashtable new_hashtable(long sz);             

/*destructor*/
void del_hashtable(hashtable ht);

#endif
