#include<int_set.h>
#include<assert.h>
#include<stdlib.h>
#include<string.h>
#include<stdio.h>

static short static_inited = 0;
static IS_WORD mask[IS_WORDBITS];
static short k2word[IS_MAXBITS];
static short k2bit[IS_MAXBITS];

static int IS_MIN(int a, int b) {if(a<=b)return a;else return b;}
static void init_static();

/***** PUBLIC *****/
int_set new_int_set(const int size){
  int_set s;
  init_static();
  s = (int_set) malloc(sizeof(struct int_set_struct));
  init_int_set(s, size);
  return s;
}

void del_int_set(int_set s){
  free(s->data);
  free(s);
}

int size_int_set(int_set s){
  return s->nelements;
}

int_set intersect(int_set A, int_set B){
  int k, nw; int_set result;
  
  result=new_int_set(IS_MIN(A->bits, B->bits));
  nw = IS_MIN(A->words, B->words);
  for(k=0;k<nw;k++)
    result->data[k]=A->data[k]&B->data[k];
  compute_nelements_int_set(result);
  return result;
}

void insert_int_set(int_set s, const int k){
  int p, r;
  
  assert(k>=0);
  if(k>=s->bits) resize_int_set(s, 2*k);
  
  p=k2word[k];
  r=k2bit[k];
  
  if(!(s->data[p] & mask[r])){
    s->data[p]|=mask[r];
    s->nelements++;
  }
}

void remove_int_set(int_set s, const int k){
  int p, r;
  assert(k>=0);
  if(k<s->bits){
    p=k2word[k];
    r=k2bit[k];
    
    if(s->data[p] & mask[r]){
      s->data[p] &= ~mask[r];
      s->nelements--;
    }
  }
}

short member_int_set(int_set s, const int k){
  int p, r;
  assert(k>=0);
  if(k>=s->bits) return 0;
  p=k2word[k]; r=k2bit[k];
  return (s->data[p]&mask[r])!=0;
}

void print_int_set(int_set s){
  int k; int first=1;
  for(k=0;k<s->bits;k++)
    if(member_int_set(s,k)) {
      if(first) {printf("%d", k); first=0;}
      else {printf(" %d", k);}
    }
  printf("\n");
}

int choose_int_set(int_set s){
  int i;
  for(i=0;i<s->bits;i++)
    if(member_int_set(s, i)) return i;
  
  return -1;
}

int_set diff(int_set A, int_set B){
  
  int_set result;
  int nA, k;

  result=new_int_set(A->bits);
  nA = A->words;
  for(k = 0; k < nA; k++){
    if(k < B->words)
      result->data[k]=A->data[k] - (A->data[k] & B->data[k]);
    else 
      result->data[k]=A->data[k];
  }
  compute_nelements_int_set(result);
  return result;
}

/*copy B into A, with B untouched*/
int_set copy_int_set(int_set A, int_set B){
  size_t bytes;

  A->bits=B->bits;
  A->words=B->words;
  A->nelements=B->nelements;
  bytes = IS_WORDBYTES*A->words;
  A->data=(IS_WORD*) realloc(A->data, bytes);
  memcpy(A->data, B->data, bytes);
  return A;
}

/***** PRIVATE *****/
void resize_int_set(int_set s, const int n){

  int old_nelements, old_words;
  size_t bytes;
  assert(n>=0 && n<IS_MAXBITS);
  
  old_nelements=s->nelements;
  old_words=s->words;

  s->bits=n; 
  s->words=s->bits/IS_WORDBITS+1;
  
  bytes=IS_WORDBYTES*s->words;
  s->data=realloc(s->data,bytes);
  assert(s->data!=0);
  
  if(s->words>old_words)
    memset(&(s->data[old_words]), '\0', (s->words-old_words)*IS_WORDBYTES);
  
  compute_nelements_int_set(s);
  assert(old_nelements==s->nelements);
}

void compute_nelements_int_set(int_set s){
  int k, i;
  s->nelements=0;
  for(k=0;k<s->words;k++)
    for(i=0;i<IS_WORDBITS;i++)
      if(s->data[k]&mask[i]) s->nelements++;
}

static void init_static(){
  int i;
  if(!static_inited){
    for(i=0;i<IS_WORDBITS;i++)
      mask[i]=(1<<i);
    for(i=0;i<IS_MAXBITS;i++){
      k2word[i]=i/IS_WORDBITS;
      k2bit[i]=i%IS_WORDBITS;
    }
    static_inited=1;
  }
}

void init_int_set(int_set s, const int maxsize){
  size_t bytes;

  assert(0<=maxsize && maxsize<=IS_MAXBITS);
  s->bits=maxsize;
  s->words=s->bits/IS_WORDBITS+1;
  s->nelements=0;
  bytes=IS_WORDBYTES*s->words;
  s->data=(IS_WORD*)malloc(bytes);
  assert(s->data!=0);
  memset(s->data,'\0',bytes);
}

