/*! \file int_set.h
\brief Contains definitions for an integer set structure
*/

#ifndef INT_SET
#define INT_SET

#define IS_WORD long
#define IS_WORDBYTES ((int)sizeof(long))
#define IS_WORDBITS (8*IS_WORDBYTES)

#define IS_MAXBITS 100000

/*! \brief An int_set is a pointer to an int_set_struct*/
typedef struct int_set_struct* int_set;

/* \brief Representation of an integer set */
struct int_set_struct {
  int bits;
  int words;
  int nelements;
  IS_WORD* data;
};

/****** PUBLIC ******/
int_set new_int_set(const int size);

void del_int_set(int_set s);

void insert_int_set(int_set s, const int k);

void remove_int_set(int_set s, const int k);

int_set intersect(int_set A, int_set B);

void resize_int_set(int_set s, const int n);

short member_int_set(int_set s, const int k);

void print_int_set(int_set s);

int size_int_set(int_set s);

/*returns smallest integer*/
int choose_int_set(int_set s);

/* returns A-B */
int_set diff(int_set A, int_set B);

/*copy B into A, with B untouched*/
int_set copy_int_set(int_set A, int_set B);

/******* PRIVATE *******/
void compute_nelements_int_set(int_set s);

void init_int_set(int_set s, const int maxsize);

#endif


