#include<phylo_tree.h>
#include<stdlib.h>
#include<stdio.h>
#include<string.h>
#include<assert.h>

phylogenetic_tree new_phylogenetic_tree(){

  int i;
  phylogenetic_tree T;
  
  T = (phylogenetic_tree) malloc(sizeof(struct phylogenetic_tree_struct));
  
  T->G = new_graph();
  T->num_leaves = 0;
  for(i = 0; i < MAX_LEAVES; i++) T->num_to_name[i] = '\0';
  return T;
}

void del_phylogenetic_tree(phylogenetic_tree T){
  
  int i;

  del_graph(T->G);
  for(i = 0; i < MAX_LEAVES; i++) 
    if(T->num_to_name[i]) free(T->num_to_name[i]);
  
  free(T);
}

void print_phylo_tree(phylogenetic_tree T) {
  
  node n; edge e;

  printf("Num of nodes = %d, num of edges = %d\n", T->G->n_nodes, T->G->n_edges);
  printf("\nEdges\n");
  for(e = T->G->first_edge; e; e = e->next){
    printf("ptr: %d id: %d wgt: %.2f u_id: %d v_id: %d\nprev: %d next: %d\nprev_u: %d next_u: %d\nprev_v: %d next_v: %d\n\n",
	   (unsigned int)e, e->id, ((edge_aux_vars)e->aux_vars)->wgt, e->u->id, e->v->id, (unsigned int)e->prev, (unsigned int)e->next, (unsigned int)e->u_prev, 
	   (unsigned int)e->u_next, (unsigned int)e->v_prev, (unsigned int)e->v_next);
  }

  for(n = T->G->first_node; n; n = n->next) {
    if(((node_aux_vars)n->aux_vars)->taxon_id != -1)
      printf("(%-3d %d %-3s %d): ", n->id, n->degree, T->num_to_name[((node_aux_vars)n->aux_vars)->taxon_id], ((node_aux_vars) n->aux_vars)->taxa_size);
    else
      printf("(%-3d %d %-3d %d): ", n->id, n->degree, ((node_aux_vars)n->aux_vars)->taxon_id, ((node_aux_vars) n->aux_vars)->taxa_size);
    
    for(e = n->first_edge; e;){
      printf(" (%d-%d):%.2f ", e->u->id, e->v->id, ((edge_aux_vars)e->aux_vars)->wgt);
      if(e->u == n) e = e->u_next;
      else if (e->v == n) e = e->v_next;
      else assert(0);
    }
    printf("\n");
  }
}

int read_tree(phylogenetic_tree T, char* newick_tree){
  
  int tree_correctly_read;
  node root, u, v, w, x;
  edge e1, e2;
  node_aux_vars node_vars;
  edge_aux_vars edge_vars;
  
  node_vars = (node_aux_vars) malloc(sizeof(struct node_aux_vars_struct));
  node_vars->taxon_id = -1;
  node_vars->taxa_size = 0;

  root = new_node(T->G, node_vars);

  tree_correctly_read = read_tree_rec(T, &newick_tree, root);
  if(!tree_correctly_read) return 0;

  /*print_phylo_tree(T);*/
  if(root->degree == 1 && ((node_aux_vars)root->aux_vars)->taxon_id == -1)
    del_node(T->G, root);
  /*print_phylo_tree(T);*/

  /*remove vertices with degree 2*/
  u = T->G->first_node;
  while(u){
    v = u->next;
    if(u->degree == 2) {
      e1 = u->first_edge; e2 = u->last_edge;
      if(e1->u != u) w = e1->u; else w = e1->v;
      if(e2->u != u) x = e2->u; else x = e2->v;

      edge_vars = (edge_aux_vars) malloc(sizeof(struct edge_aux_vars_struct));
      if(e1->aux_vars && e2->aux_vars)
	edge_vars->wgt = ((edge_aux_vars)e1->aux_vars)->wgt 
	  + ((edge_aux_vars)e2->aux_vars)->wgt;
      else 
	edge_vars->wgt = 0;

      new_edge(T->G, w, x, edge_vars);
      del_node(T->G, u);
    }
    u = v;
  }

  /*print_phylo_tree(T);*/
  return 1;
}
	
int read_tree_rec(phylogenetic_tree T, char** newick_tree, node root){
  
  node n;
  char* endptr;
  double x;
  node_aux_vars node_vars;
  edge_aux_vars edge_vars;

  node_vars = (node_aux_vars) malloc(sizeof(struct node_aux_vars_struct));
  node_vars->taxon_id = -1;
  node_vars->taxa_size = 0;
  
  n = new_node(T->G, node_vars);

  if(**newick_tree=='('){
    /*recurse to read rest of tree*/
    do {
      (*newick_tree)++;
      if(!read_tree_rec(T, newick_tree, n)) return 0;
    }
    while(**newick_tree == ',');
    
    if(**newick_tree != ')'){
      printf("Error: expected: ')', found: '%c'\n", **newick_tree);
      return 0;
    }
    (*newick_tree)++;
  }
  else { 
    /*found leaf; read in the name*/
    if(!read_leaf_name(T, newick_tree, n)){
      printf("Error: expected taxon name before: %c\n", **newick_tree);
      exit(EXIT_FAILURE);
    }
  }

  if(**newick_tree == ':') {
    /*found weight; read in weight*/
    (*newick_tree)++;
    x = strtod(*newick_tree, &endptr);
    if(*newick_tree == endptr) {
      printf("Error: expected double, found %c\n", **newick_tree);
      exit(EXIT_FAILURE);
    }
    *newick_tree = endptr;

    edge_vars = (edge_aux_vars) malloc(sizeof(struct edge_aux_vars_struct));
    edge_vars->wgt = x;
    new_edge(T->G, root, n, edge_vars);
    ((node_aux_vars)root->aux_vars)->taxa_size += ((node_aux_vars)n->aux_vars)->taxa_size;
  }
  else {
    edge_vars = (edge_aux_vars) malloc(sizeof(struct edge_aux_vars_struct));
    edge_vars->wgt = 0;
    new_edge(T->G, root, n, edge_vars);
    ((node_aux_vars)root->aux_vars)->taxa_size += ((node_aux_vars)n->aux_vars)->taxa_size;
  }
  
  return 1;
}

int read_leaf_name(phylogenetic_tree T, char** newick_tree, node n){
  
  char name[MAX_NAME_LEN];
  int i = 0;

  while(!strchr(":(),", **newick_tree)){
    name[i++] = **newick_tree;
    (*newick_tree)++;
    if(!*newick_tree){
      printf("Error: unexpected EOF while reading leaf name\n");
      return 0;
    }
  }
  name[i] = '\0';

  ((node_aux_vars)n->aux_vars)->taxon_id = T->num_leaves;
  ((node_aux_vars)n->aux_vars)->taxa_size = 1;

  T->num_to_name[T->num_leaves] = (char*) malloc(strlen(name));
  strcpy(T->num_to_name[T->num_leaves], name);
  T->num_leaves++;
  return 1;
}


