#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<tree2splits.h>
#include<splits2tree.h>
#include<constants.h>
#include<phylo_tree.h>
#include<hashtable.h>
#include<assert.h>
#include<list.h>
#include<split.h>
#include<scm.h>
#include<ugraph_aux_vars.h>

int VERBOSE=0;

void read_chars(FILE* fp, char*** num_to_names, char*** sequences, 
		int* ntax, int* nchars);

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char* out_tree;
  FILE* fp; int i, j, t; 
  phylogenetic_tree* tree_array;
  list* list_array;
  int_set* taxa_array;
  char** num_to_name;
  char** sequences;
  int ntax, nchars;
  hashtable name_to_num;
  int num_trees;
  node v;
  list_node ln, ln2;
  list scm_splits;
  phylogenetic_tree scm_tree;
  int_set X, A, B, taxa;
  split S; 
  int* intset_entries;

  tree_array = (phylogenetic_tree*) malloc(2*sizeof(phylogenetic_tree));
  list_array = (list*) malloc(2*sizeof(list));
  taxa_array = (int_set*) malloc(2*sizeof(int_set));

  num_trees = 0;

  if(argc < 2){
    printf("Please specify guide tree using \"-gt <filename containing guide_tree>\"\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-s")){
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      read_chars(fp, &num_to_name, &sequences, &ntax, &nchars);
      assert(num_to_name); assert(sequences);
      name_to_num = new_hashtable(ntax);
      for(j=0;j<ntax;j++){
	insert_hashtable(name_to_num, num_to_name[j], j);
      }
      fclose(fp);
    }
    /*read in the newick tree*/
    else if (!strcmp(argv[i], "-gt")) {
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      j = 0;
      while(fscanf(fp, "%s", newick_tree) != EOF){
	assert(name_to_num);
	tree_array[j] = new_phylogenetic_tree();
	read_tree(tree_array[j], num_to_name, newick_tree, name_to_num);
	/*
	out_tree=(char*)malloc(MAX_NEWICK_STR_LEN*sizeof(char));
	print_tree(tree_array[j], out_tree, 1);
	printf("read tree:%s\n", out_tree);
	free(out_tree);
	*/
	j++;
      }
      num_trees = j;
      fclose(fp);
    }
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%d", &VERBOSE);
    }
    else {
      printf("Please specify dataset using \"-s <filename containing sequences in phylip format>\"\n");
    }
  }

  for(i=0;i<num_trees;i++){
    taxa_array[i] = new_int_set(32);
    for(v = tree_array[i]->G->first_node; v; v=v->next){
      if(v->degree == 1){
	insert_int_set(taxa_array[i], ((node_aux_vars)v->aux_vars)->taxon_id);
      }
    }
  }

  printf("doing tree2splits\n");
  for(i=0;i<num_trees;i++){
    list_array[i] = new_list();
    tree2splits(tree_array[i], list_array[i]);
  }
  printf("ending tree2splits\n");
  
  scm_splits = new_list(); 
  X = intersect_int_set(taxa_array[0],taxa_array[1]);

  printf("doing scm\n");
  scm(taxa_array[0], taxa_array[1], X, list_array[0], list_array[1],
      scm_splits);
  printf("out of scm\n");
  del_int_set(X);
  
  X = union_int_set(taxa_array[0], taxa_array[1]);
  assign_int_set(taxa_array[0], X);
  del_int_set(X);

  taxa=new_int_set(32);
  assign_int_set(taxa, taxa_array[0]);

  /*add trivial splits*/
  for(ln=scm_splits->head; ln; ln=ln->next){
    A = get_A_split((split)ln->data);
    if(size_int_set(A)==1)
      remove_int_set(taxa_array[0], choose_int_set(A));
    else if (size_int_set(A)==size_int_set(taxa)-1){
      /*B=get_B_split((split)ln->data, taxa);*/
      B=get_B_split((split)ln->data);
      remove_int_set(taxa_array[0], choose_int_set(B));
    }
  }
  intset_entries = malloc(size_int_set(taxa_array[0]) * sizeof(int));
  get_entries_int_set(taxa_array[0], intset_entries);
  for(i=0;i<size_int_set(taxa_array[0]);i++){
    t = intset_entries[i];
    S = new_split();
    B = new_int_set(32);
    assign_int_set(B, taxa);
    A = new_int_set(32);
    insert_int_set(A, t);
    remove_int_set(B, t);
    set_split(S, A, B, 1);
    del_int_set(A); del_int_set(B);
    append_list(scm_splits, S);
  }
  free(intset_entries);
  /*********************/

  del_int_set(taxa_array[0]); del_int_set(taxa_array[1]);
  for(ln=list_array[0]->head; ln; ln=ln->next){
    del_split((split)ln->data); ln->data=NULL;
  }
  for(ln=list_array[1]->head; ln; ln=ln->next){
    del_split((split)ln->data); ln->data=NULL;
  }
  del_list(list_array[0]); del_list(list_array[1]);

  scm_tree = new_phylogenetic_tree();

  printf("calling splits2tree\n");
  splits2tree(scm_splits, taxa, scm_tree);
  printf("done from splits2tree\n");
  
  printf("scm tree has %d nodes and %d edges\n", scm_tree->G->n_nodes,
	 scm_tree->G->n_edges);

  out_tree=(char*)malloc(MAX_NEWICK_STR_LEN*sizeof(char));
  for(t=0;t<MAX_NEWICK_STR_LEN;t++) out_tree[t] = '\0';

  printf("printing tree\n");
  print_tree(scm_tree, num_to_name, out_tree, 1);
  printf("scm tree %s\n", out_tree);
  free(out_tree);
}

void read_chars(FILE* fp, char*** num_to_names_, char*** sequences_,
		int* ntax, int* nchars){
  
  int i; char** sequences; char** num_to_names;

  *ntax=0; *nchars=0;
  fscanf(fp, "%d", ntax);
  fscanf(fp, "%d", nchars);
  if(*ntax<=0 || *nchars<=0){
    printf("Error: number of taxa=%d and number of sites=%d must be positive", *ntax, *nchars);
    exit(EXIT_FAILURE);
  }
  
  sequences=(char**) malloc(*ntax * sizeof(char*));
  num_to_names=(char**) malloc(*ntax * sizeof(char*));
  for(i=0;i<*ntax;i++){
    sequences[i] = (char*) malloc((*nchars+1) * sizeof(char));
    num_to_names[i] = (char*) malloc(MAX_NAME_LEN * sizeof(char));
    fscanf(fp, "%s", num_to_names[i]);
    fscanf(fp, "%s", sequences[i]);
  }

  /*
  for(i=0;i<*ntax;i++)
    printf("%s\t%s\n", num_to_names[i], sequences[i]);
  */

  *sequences_ = sequences;
  *num_to_names_ = num_to_names;
}

