#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<tree2splits.h>
#include<splits2tree.h>
#include<constants.h>
#include<phylo_tree.h>
#include<split.h>
#include<ugraph_aux_vars.h>
#include<dcm3.h>

int VERBOSE=0;

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char newick_tree2[MAX_NEWICK_STR_LEN];
  char newick_tree3[MAX_NEWICK_STR_LEN];
  char newick_tree4[MAX_NEWICK_STR_LEN];
  FILE* fp; int i, j; 
  phylogenetic_tree T, U;
  list listsplits; int_set taxa; list_node ln;
  node v; edge e; edge c_edge;
  char* num_to_name[MAX_LEAVES];
  int* subsets; int* subsets_; int* separator;
  int subset_num, separator_size;

  if(argc < 2){
    printf("Please specify guide tree using \"-gt <filename containing guide_tree>\"\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in the newick tree*/
    if (!strcmp(argv[i], "-gt")) {
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      fscanf(fp, "%s", newick_tree);
      fclose(fp);
    }
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%d", &VERBOSE);
    }
    else {
      printf("Please specify guide tree using \"-gt <filename containing guide_tree>\"\n");
    }
  }
  
  T = new_phylogenetic_tree();
  if(!read_tree(T, num_to_name, newick_tree, 0)) {
    printf("Error: could not read tree %s\n", newick_tree);
    exit(EXIT_FAILURE);
  }

  print_tree(T, num_to_name, newick_tree2, 1);
  printf("T = %s\n", newick_tree2);

  U = new_phylogenetic_tree();
  printf("copying....\n");
  assign_tree(U, T);
  printf("copying done!\n");

  print_phylo_tree(T, num_to_name);
  del_phylogenetic_tree(T);

  set_subtree_sizes(U);
  print_tree(U, num_to_name, newick_tree3, 1);
  printf("U = %s\n", newick_tree3);
  print_phylo_tree(U, num_to_name);

  taxa = new_int_set(32);
  insert_int_set(taxa, 0);
  insert_int_set(taxa, 2);
  insert_int_set(taxa, 5);
  insert_int_set(taxa, 6);
  insert_int_set(taxa, 8);
  insert_int_set(taxa, 10);
  
  induce_tree(U, taxa);
  print_tree(U, num_to_name, newick_tree4, 1);
  printf("U = %s\n", newick_tree4);

  set_subtree_sizes(U);
  print_phylo_tree(U, num_to_name);

  subsets = (int*) malloc(U->num_leaves * sizeof(int));
  subsets_ = (int*) malloc(U->num_leaves * sizeof(int));
  separator = (int*) malloc(U->num_leaves * sizeof(int));

  c_edge = get_centroid_edge(U);
  printf("Tree has leaves %d\n", U->num_leaves);
  printf("Centroid edge = (%d-%d)\n", c_edge->u->id, c_edge->v->id);

  get_dcm3_subsets(U, c_edge, subsets, subsets_, separator,
		   &subset_num);
  
  printf("num of subsets=%d\n", subset_num-1);

  for(i = 0; i < subset_num-1;i++){
    for(j = subsets_[i]; j < subsets_[i+1]; j++){
      printf("%s\t", num_to_name[subsets[j]]);
    }
    for(j = 0; separator[j] != -1 && j < U->num_leaves;j++){
      printf("%s\t", num_to_name[separator[j]]);
    }
    printf("\n");
  }

}
