/*
This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<tree2splits.h>
#include<ugraph_aux_vars.h>
#include<split.h>
#include<assert.h>

int VERBOSE=0;
int MAX_FPFN_TREES=10;

void falseposneg(const list model_splits, const list estimated_splits, float t, int* fp, int* fn, int* cm, int* num_msplits, int* num_esplits);

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char* num_to_name[MAX_LEAVES];

  int i, ntax, j, fp, fn, common, num_msplits, num_esplits;
  float threshold = 0.5;
  phylogenetic_tree model_tree;
  phylogenetic_tree trees[MAX_FPFN_TREES];
  int_set taxa; node n;
  hashtable name_to_num;
  list model_splits;
  list splits[MAX_FPFN_TREES];

  /*initialization*/

  if(argc < 2){
    printf("Type fpfn -help on how to run\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-t")){
      i++;
      sscanf(argv[i], "%f", &threshold);
      assert(threshold >= 0 && threshold <= 0.5);
    }
    else if (!strcmp(argv[i], "-help")){
      printf("cat modeltree tree1 tree2 ... treen | fpfn -t <threshold>\n");
      exit(EXIT_FAILURE);
    }
    else {
      printf("Type fpfn -help on how to run\n");
      exit(EXIT_FAILURE);
    }
  }

  /*read model tree*/
  if(!scanf("%s",newick_tree)){
    printf("Error: no model tree found\n");
    exit(EXIT_FAILURE);
  }
  else {
    model_tree = new_phylogenetic_tree();
    if(!read_tree(model_tree, num_to_name, newick_tree, 0)) {
      printf("Error: could not read tree %s\n", newick_tree);
      exit(EXIT_FAILURE);
    }
  }

  ntax = model_tree->num_leaves;
  /*get name to num hashtables */
  name_to_num = new_hashtable(ntax);
  for(j=0;j<ntax;j++){
    /*checking for duplicate taxon names*/
    assert(lookup_hashtable(name_to_num,num_to_name[j])==-1);
    insert_hashtable(name_to_num, num_to_name[j], j);
  }

  /*read remaining trees*/
  i=0;
  while(scanf("%s", newick_tree) != EOF){
    trees[i] = new_phylogenetic_tree();
    if(!read_tree(trees[i], num_to_name, newick_tree, name_to_num)) {
      printf("Error: could not read tree %s\n", newick_tree);
      exit(EXIT_FAILURE);
    }

    /*checking for duplicate taxon names*/
    taxa=new_int_set(32);
    for(n=trees[i]->G->first_node;n;n=n->next)
      if(n->degree==1 && ((node_aux_vars)n->aux_vars)->taxon_id!=-1)
        insert_int_set(taxa,((node_aux_vars)n->aux_vars)->taxon_id);
    if(size_int_set(taxa)!=ntax){
      printf("Error: duplicate taxon names in estimated tree\n");
      exit(EXIT_FAILURE);
    }

    /*make sure ntax is equal to model tree ntax*/
    if(ntax != trees[i]->num_leaves){
      printf("Error: ntax in model and estimated tree not equal\n");
      exit(EXIT_FAILURE);
    } 
    i++;
  }
  if(i==0){
    printf("Error: no trees to compare against model tree\n");
    exit(EXIT_FAILURE);
  }

  /*convert each tree to split*/
  model_splits = new_list();
  tree2splits(model_tree, model_splits);
  for(j=0;j<i;j++){
    splits[j]=new_list();
    tree2splits(trees[j], splits[j]);
  }

  fp=-1; fn=-1; common=-1;
  for(j=0;j<i;j++){
    falseposneg(model_splits,splits[j],threshold,&fp,&fn,&common,&num_msplits,&num_esplits);
    printf("%d %d %d %d %d\n",fp,fn,common,num_msplits,num_esplits);
  }
  
  return 1;
}

void falseposneg(const list model_splits, const list estimated_splits, float t, int* fp, int* fn, int* cm, int* num_msplits, int* num_esplits){

  list_node it1; list_node it2; split sp1; split sp2;
  int_set s1; int_set s2;
  int common, i1, i2, ntax, t1, t2; 
  ntax = size_int_set(get_A_split((split)model_splits->head->data)) +
         size_int_set(get_B_split((split)model_splits->head->data));
  common=0; t1=0; t2=0;

  i1=0;
  for(it1=model_splits->head; it1; it1=it1->next){
    sp1 = (split)it1->data; s1=get_A_split(sp1);
    if(size_int_set(s1)==1 || size_int_set(s1)==ntax-1) continue;
    if(size_int_set(s1) <= t*ntax || (ntax - size_int_set(s1)) <= t*ntax){
      t1++; i1++;
      //if(size_int_set(s1)==1 || size_int_set(s1)==ntax-1) continue; else i1++;
      //printf("%d <= %.2f\n", size_int_set(s1), t*ntax);
      for(it2=estimated_splits->head; it2; it2=it2->next){
        sp2 = (split)it2->data; s2=get_A_split(sp2);
        if(size_int_set(s2)==1 || size_int_set(s2)==ntax-1) continue; 
        if(size_int_set(s2) <= t*ntax || (ntax - size_int_set(s2)) <= t*ntax)
          if(equal_int_set(s1,s2)){common++; i1--; break;}
      }
    }
  }

  i2=0;
  for(it1=estimated_splits->head; it1; it1=it1->next){
    sp1 = (split)it1->data; s1=get_A_split(sp1);
    if(size_int_set(s1)==1 || size_int_set(s1)==ntax-1) continue;
    if(size_int_set(s1) <= t*ntax || (ntax - size_int_set(s1)) <= t*ntax){
      t2++; i2++;
      //if(size_int_set(s1)==1 || size_int_set(s1)==ntax-1) continue; else i2++;
      for(it2=model_splits->head; it2; it2=it2->next){
        sp2 = (split)it2->data; s2=get_A_split(sp2);
        if(size_int_set(s2)==1 || size_int_set(s2)==ntax-1) continue; 
        if(size_int_set(s2) <= t*ntax || (ntax - size_int_set(s2)) <= t*ntax)
          if(equal_int_set(s1,s2)){i2--; break;}
      }
    }
  }


  *fp=i2; *fn=i1; *cm=common; *num_msplits=t1; *num_esplits=t2;
  return;
}


