/*
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<assert.h>

int VERBOSE=0;

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char newick_tree2[MAX_NEWICK_STR_LEN];
  char* num_to_name[MAX_LEAVES];
  char command[MAX_BUFFER];
  FILE* fp; int i; 
  phylogenetic_tree T;

  /**Initialization**/
  T = NULL;
  
  if(argc < 2){
    printf("Type treefunctions -help on how to run\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in the newick tree*/
    if (!strcmp(argv[i], "-t")) {
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      fscanf(fp, "%s", newick_tree);
      fclose(fp);

      T = new_phylogenetic_tree();
      if(!read_tree(T, num_to_name, newick_tree, 0)) {
	printf("Error: could not read tree %s\n", newick_tree);
	exit(EXIT_FAILURE);
      }
    }
    else if (!strcmp(argv[i], "-c")){
      i++; sscanf(argv[i], "%s", command);
      assert(command[0] != '-');
    }
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%d", &VERBOSE);
    }
    else if (!strcmp(argv[i], "-help")){
      printf("Usage: treefunctions\n\t-t <filename containing newick tree> (tree must NOT contain any spaces)\n\t-c <command> (command can be one of the following)\n\trandomrefine_edgelens: randomly refine the input tree and assign 0 edge length to new edges\n\tinternal: return number of internal edges\n\tntax return number of taxa in tree\n");
      exit(EXIT_FAILURE);
    }
    else {
      printf("Type treefunctions -help on how to run\n");
      exit(EXIT_FAILURE);
    }
  }
  
  if(!strcmp(command, "randomrefine_edgelens")){
    random_refine_edgelens(T);
    for(i=0;i<strlen(newick_tree2);i++)newick_tree2[i]='\0';
    print_tree(T, num_to_name, newick_tree2, 1);
    printf("%s\n", newick_tree2);
    
    for(i=0;i<T->num_leaves;i++)free(num_to_name[i]);
    del_phylogenetic_tree(T);
  }
  else if (!strcmp(command, "internal")){
    printf("%d\n", T->G->n_edges - T->num_leaves);
  }
  else if (!strcmp(command, "ntax")){
    printf("%d\n", T->num_leaves);
  }
  return 1;
}
