/*
This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<tree2splits.h>
#include<ugraph_aux_vars.h>
#include<split.h>
#include<assert.h>
#include<int_set.h>

int VERBOSE=0;
int MAX_FPFN_TREES=10;

void falseposneg(const list model_splits, const list estimated_splits, float t, int* fp, int* fn, int* cm, int* num_msplits, int* num_esplits);

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char* num_to_name[MAX_LEAVES];

  int i, ntax, j, fp, fn, common, num_msplits, num_esplits;
  float threshold = 0.5;
  phylogenetic_tree model_tree;
  phylogenetic_tree trees[MAX_FPFN_TREES];
  int_set taxa; node n;
  hashtable name_to_num;
  list model_splits;
  list splits[MAX_FPFN_TREES];

  /*initialization*/

  /*
  if (!strcmp(argv[1], "-help")){
      printf("cat modeltree tree1 tree2 ... treen | fpfn -t <threshold>\n");
      exit(EXIT_FAILURE);
  }
  */

  /*read model tree*/
  if(!scanf("%s",newick_tree)){
    printf("Error: no model tree found\n");
    exit(EXIT_FAILURE);
  }
  else {
    model_tree = new_phylogenetic_tree();
    if(!read_tree(model_tree, num_to_name, newick_tree, 0)) {
      printf("Error: could not read tree %s\n", newick_tree);
      exit(EXIT_FAILURE);
    }
  }

  ntax = model_tree->num_leaves;
  /*get name to num hashtables */
  name_to_num = new_hashtable(ntax);
  for(j=0;j<ntax;j++){
    /*checking for duplicate taxon names*/
    assert(lookup_hashtable(name_to_num,num_to_name[j])==-1);
    insert_hashtable(name_to_num, num_to_name[j], j);
  }

  /*convert tree to splits and print splits*/
  model_splits = new_list();
  tree2splits(model_tree, model_splits);
  //printf("size of model_splits=%d\n", size_list(model_splits));
  list_node it;
  for(it=model_splits->head; it; it=it->next){
    split s = (split) it->data;
    //split s = (split) pop_list(model_splits);
    int_set is;
    if(size_int_set(get_A_split(s)) < size_int_set(get_B_split(s))){
      is = get_A_split(s);
    }
    else {
      is = get_B_split(s);
    }
    if(size_int_set(is) > 1) { print_int_set2(is, num_to_name); printf("\n"); }
  }
  
  return 1;
}

