/*
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<assert.h>
#include<ugraph_aux_vars.h>

int VERBOSE=0;

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char newick_tree2[MAX_NEWICK_STR_LEN];
  char* num_to_name[MAX_LEAVES];
  /*char command[MAX_BUFFER];*/

  edge edgearray[10];

  FILE* fp; int i, j, k, dir; phylogenetic_tree T;
  edge e,f,e1,e2; node u,w,x; 
  edge_aux_vars evars; 
  /*node_aux_vars nvars;*/
  
  /**Initialization**/
  T = NULL;
  dir = -1; /*0==left, 1==right*/

  if(argc < 2){
    printf("Type rerootree -help on how to run\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in the newick tree and edge to reroot*/
    if (!strcmp(argv[i], "-t")) {
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      assert(fscanf(fp, "%s", newick_tree));
      fclose(fp);

      T = new_phylogenetic_tree();
      if(!read_rooted_tree(T, num_to_name, newick_tree, 0)) {
	printf("Error: could not read tree %s\n", newick_tree);
	exit(EXIT_FAILURE);
      }
    }
    /*else if (!strcmp(argv[i], "-d")){
      i++;
      sscanf(argv[i], "%d", &dir);
    }*/
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%d", &VERBOSE);
    }
    else if (!strcmp(argv[i], "-help")){
      printf("To do....\n");
      exit(EXIT_FAILURE);
    }
    else {
      printf("Type rerootree -help on how to run\n");
      exit(EXIT_FAILURE);
    }
  }
  
  /*  
  assert(0 <= dir && dir <= 3);
  if(edge_num > T->G->n_edges) { 0
    printf("Error: edge_num=%d > edges in T=%d\n", edge_num, T->G->n_edges);
  }
  */

  for(u = T->G->first_node; u; u=u->next){
    if(u->degree == 2){

      /*delete node u*/
      assert(((node_aux_vars)u->aux_vars)->taxon_id == -1);
      e1 = u->first_edge; e2 = u->last_edge;
      if(e1->u != u) w = e1->u; else w = e1->v;
      if(e2->u != u) x = e2->u; else x = e2->v;
        
      evars = new_edge_aux_vars();
      if(e1->aux_vars && e2->aux_vars)
        evars->wgt = ((edge_aux_vars)e1->aux_vars)->wgt 
                       + ((edge_aux_vars)e2->aux_vars)->wgt;
      else 
        evars->wgt = 0;
        
      e = new_edge(T->G, w, x, evars);
      del_node(T->G, u); u = NULL;

      w=e->v; i=0; if(w->degree==1) printf("Warning: no subtrees on one side\n");
      for(f=w->first_edge; f;){
        if(f!=e) edgearray[i++]=f;
        if(f->u==w)f=f->u_next;else if(f->v==w)f=f->v_next; else assert(0);     
      }

      w=e->u; if(w->degree==1) printf("Warning: no subtrees on one side\n");
      for(f=w->first_edge; f;){
        if(f!=e) edgearray[i++]=f;
        if(f->u==w)f=f->u_next;else if(f->v==w)f=f->v_next; else assert(0);     
      }

      for(j=0;j<i;j++){
        assert(edgearray[j]);
        reroot_phylo_tree(T, edgearray[j]);

        /*print tree*/
        for(k=0;k<strlen(newick_tree2);k++)newick_tree2[k]='\0';
        print_rooted_tree(T, num_to_name, newick_tree2, 1);
        printf("%s\n", newick_tree2);
      }
     
      break;
    }
  }

  return 1;
}

