/*
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<assert.h>
#include<ugraph_aux_vars.h>

int VERBOSE=0;

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char newick_tree2[MAX_NEWICK_STR_LEN];
  char* num_to_name[MAX_LEAVES];
  /*char command[MAX_BUFFER];*/
  edge eu[2]; edge ev[2];

  FILE* fp; int i, j, k, edge_num; 
  phylogenetic_tree T;
  edge e, f, n1, n2;
  node u, v, ou, ov;
  edge_aux_vars evars;

  /**Initialization**/
  T = NULL;
  edge_num = -1;

  if(argc < 2){
    printf("Type nni -help on how to run\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in the newick tree and edge number to perform NNI*/
    if (!strcmp(argv[i], "-t")) {
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      assert(fscanf(fp, "%s", newick_tree));
      fclose(fp);

      T = new_phylogenetic_tree();
      if(!read_rooted_tree(T, num_to_name, newick_tree, 0)) {
	printf("Error: could not read tree %s\n", newick_tree);
	exit(EXIT_FAILURE);
      }
    }
    else if (!strcmp(argv[i], "-e")){
      i++;
      sscanf(argv[i], "%d", &edge_num);
      assert(edge_num >= 0);
    }
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%d", &VERBOSE);
    }
    else if (!strcmp(argv[i], "-help")){
      printf("To do....\n");
      exit(EXIT_FAILURE);
    }
    else {
      printf("Type nni -help on how to run\n");
      exit(EXIT_FAILURE);
    }
  }
  
  if(edge_num > T->G->n_edges) { 
    printf("Error: edge_num=%d > edges in T=%d\n", edge_num, T->G->n_edges);
  }

  set_subtree_edgelens(T);

  i=0;
  for(e = T->G->first_edge; e; e=e->next){
    if(e->u->degree>2 && e->v->degree>2){
      if(i==edge_num){
	/*perform NNI on this edge and output trees. 
	  don't care about edgelens for now.*/
        printf("NNI edge has weight %.3f\n", ((edge_aux_vars)e->aux_vars)->wgt);
        printf("NNI edge has weight %.3f\n", ((edge_aux_vars)e->aux_vars)->wgt);

	u = e->u; v = e->v;
	
	j=0;
	for(f=u->first_edge; f;) { 
	  if(f != e) eu[j++] = f; 
	  if(f->u==u)f=f->u_next;else if(f->v==u)f=f->v_next; else assert(0);
	}
	assert(j==2);
	
	j=0;
	for(f=v->first_edge; f;) { 
	  if(f != e) ev[j++] = f; 
	  if(f->u==v)f=f->u_next;else if(f->v==v)f=f->v_next; else assert(0);
	}
	assert(j==2);
	
	for(j=0;j<2;j++){
	  /*remove old edges and add new edges*/
	  ov=opposite(v, ev[0]);
	  ou=opposite(u, eu[j]);
	  
	  evars=new_edge_aux_vars();evars->wgt=1;
	  n1 = new_edge(T->G, ou, v, evars);
	  
	  evars=new_edge_aux_vars();evars->wgt=1;	
	  n2 = new_edge(T->G, ov, u, evars);
	  
	  del_edge(T->G, ev[0]);
	  del_edge(T->G, eu[j]);
	  
	  /*print tree*/
	  for(k=0;k<strlen(newick_tree2);k++)newick_tree2[k]='\0';
	  print_rooted_tree(T, num_to_name, newick_tree2, 1);
	  printf("%s\n", newick_tree2);
	  
	  /*restore tree*/
	  del_edge(T->G, n1);
	  del_edge(T->G, n2);
	  
	  evars=new_edge_aux_vars();evars->wgt=1;
	  ev[0] = new_edge(T->G, ov, v, evars);
	  
	  evars=new_edge_aux_vars();evars->wgt=1;
	  eu[j] = new_edge(T->G, ou, u, evars);
	}
      }
      i++;
    }
  }

  return 1;
}


