/*
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include <time.h>
#include<constants.h>
#include<phylo_tree.h>
#include<assert.h>
#include<ugraph_aux_vars.h>

typedef struct pair_struct* pair;
struct pair_struct{
  node u; node v;
};

static void remove_2degree_butnotroot_tree(phylogenetic_tree T, const node root);
static void permutearray(int* p, int size);
static void get_edges(list q_parent, list q_child, const phylogenetic_tree T, list l);
static int min(int x, int y);

int VERBOSE=0;

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char newick_tree2[MAX_NEWICK_STR_LEN];
  char* num_to_name[MAX_LEAVES];
  /*char command[MAX_BUFFER];*/
  edge eu[2]; edge ev[2];
  node nu[2]; node nv[2];

  FILE* fp; int i, j, k, m, edge_num, tbr_trees; 
  phylogenetic_tree T;
  edge e, f, n1, n2, d1, d2, ne;
  node u, v, ou, ov, lu, lv, ru, rv, u_, v_, euu, euv, evu, evv;
  edge_aux_vars evars;
  node_aux_vars nvars;
  int *p1, *p2;
  list edgelistL; list edgelistR; list q_parent; list q_child;
  list_node ln;
  pair pair1; pair pair2;

  /**Initialization**/
  T = NULL;
  edge_num = -1;
  tbr_trees=1;

  /*initialize random number seed  */
  srand(time(NULL));

  if(argc < 2){
    printf("Type nni -help on how to run\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in the newick tree and edge number to perform NNI*/
    if (!strcmp(argv[i], "-t")) {
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      assert(fscanf(fp, "%s", newick_tree));
      fclose(fp);

      T = new_phylogenetic_tree();
      if(!read_rooted_tree(T, num_to_name, newick_tree, 0)) {
	printf("Error: could not read tree %s\n", newick_tree);
	exit(EXIT_FAILURE);
      }
    }
    else if (!strcmp(argv[i], "-e")){
      i++;
      sscanf(argv[i], "%d", &edge_num);
      assert(edge_num >= 0);
    }
    else if (!strcmp(argv[i], "-n")){
      i++;
      sscanf(argv[i], "%d", &tbr_trees);
      assert(edge_num >= 0);
    }
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%d", &VERBOSE);
    }
    else if (!strcmp(argv[i], "-help")){
      printf("To do....\n");
      exit(EXIT_FAILURE);
    }
    else {
      printf("Type nni -help on how to run\n");
      exit(EXIT_FAILURE);
    }
  }
  
  if(edge_num > T->G->n_edges) { 
    printf("Error: edge_num=%d > edges in T=%d\n", edge_num, T->G->n_edges);
    exit(EXIT_FAILURE);
  }

  set_subtree_edgelens(T);

  i=0;
  for(e = T->G->first_edge; e; e=e->next){
    if(e->u->degree>2 && e->v->degree>2){
      if(i==edge_num){

	/*perform random TBR on this edge and output trees. 
	  don't care about edgelens for now.*/
        printf("TBR edge has weight %.3f\n", ((edge_aux_vars)e->aux_vars)->wgt);

	u = e->u; v = e->v;
	
	edgelistL=new_list(); edgelistR=new_list();
	q_parent=new_list(); q_child=new_list();
	append_list(q_parent, (void *) v);
	append_list(q_child, (void *) u);
	get_edges(q_parent, q_child, T, edgelistL);

	assert(!size_list(q_parent) && !size_list(q_child));
	append_list(q_parent, (void *) u);
	append_list(q_child, (void *) v);
	get_edges(q_parent, q_child, T, edgelistR);
	del_list(q_parent); del_list(q_child);

	/*printf("Edges in left edge list=%d\n", size_list(edgelistL));
	  printf("Edges in right edge list=%d\n", size_list(edgelistR));*/

	/*randomly permute*/
	p1 = malloc(size_list(edgelistL)*sizeof(int));
	p2 = malloc(size_list(edgelistR)*sizeof(int));
	for(j=0;j<size_list(edgelistL);j++)p1[j]=j;
	for(j=0;j<size_list(edgelistR);j++)p2[j]=j;
	/*permutearray(p1, size_list(edgelistL));
	  permutearray(p2, size_list(edgelistR));*/

	if(size_list(edgelistL)<tbr_trees)tbr_trees=size_list(edgelistL);
	if(size_list(edgelistR)<tbr_trees)tbr_trees=size_list(edgelistR);

	for(j=2;j<min(size_list(edgelistL),5);j++){
	  for(m=2;m<min(size_list(edgelistR),5);m++){
	    if(j==1 || m==1 || (m==0 && j==0)) continue;

	    /*get pairs*/
	    k=0; for(ln = edgelistL->head; ln; ln = ln->next){
	      /*if(k==p1[j]) {n1=ln->data; break;} k++;*/
	      if(k==p1[j]) {pair1=ln->data; break;} k++;
	    }
	    k=0; for(ln = edgelistR->head; ln; ln = ln->next){
	      /*if(k==p2[m]) {n2=ln->data; break;} k++;*/
	      if(k==p2[m]) {pair2=ln->data; break;} k++;
	    }
	    
	    /*get edges from pairs*/
	    for(f=T->G->first_edge;f;f=f->next)
	      if((f->u==pair1->u || f->v==pair1->u) && (f->u==pair1->v || f->v==pair1->v)){
		n1=f; break;
	      }
	    for(f=T->G->first_edge;f;f=f->next)
	      if((f->u==pair2->u || f->v==pair2->u) && (f->u==pair2->v || f->v==pair2->v)){
		n2=f; break;
	      }
	    
	    printf("connecting edges %d and %d\n", p1[j], p2[m]);
	    
	    k=0;
	    for(f=u->first_edge; f;) { 
	      if(f != e){ nu[k] = opposite(u, f); eu[k] = f; k++;}
	      if(f->u==u)f=f->u_next;else if(f->v==u)f=f->v_next; else assert(0);
	    }
	    assert(k==2); k=0;
	    for(f=v->first_edge; f;) { 
	      if(f != e){ nv[k] = opposite(v, f); ev[k] = f; k++;}
	      if(f->u==v)f=f->u_next;else if(f->v==v)f=f->v_next; else assert(0);
	    }
	    assert(k==2);
	    /*delete bisection edge e*/
	    del_edge(T->G, e);
	    
	    /*bisect each edge and connect*/
	    lu=n1->u; lv=n1->v; 
	    if(lu==u || lv==u){u_=u;}
	    else{
	      nvars = new_node_aux_vars(); u_=new_node(T->G, nvars);
	      evars = new_edge_aux_vars(); new_edge(T->G, lu, u_, evars);
	      evars = new_edge_aux_vars(); new_edge(T->G, lv, u_, evars);
	      del_edge(T->G, n1);
	      evars = new_edge_aux_vars(); d1=new_edge(T->G, nu[0], nu[1], evars);
	      assert(eu[0]!=n1 && eu[1]!=n1); del_edge(T->G,eu[0]); del_edge(T->G,eu[1]);
	    }
	    
	    ru=n2->u; rv=n2->v; 
	    if(ru==v || rv==v){v_=v;}
	    else{
	      nvars = new_node_aux_vars(); v_=new_node(T->G, nvars);
	      evars = new_edge_aux_vars(); new_edge(T->G, ru, v_, evars);
	      evars = new_edge_aux_vars(); new_edge(T->G, rv, v_, evars);
	      del_edge(T->G, n2);
	      evars = new_edge_aux_vars(); d2=new_edge(T->G, nv[0], nv[1], evars);
	      assert(ev[0]!=n2 && ev[1]!=n2); del_edge(T->G,ev[0]);del_edge(T->G,ev[1]);
	    }
	    evars = new_edge_aux_vars(); ne=new_edge(T->G, u_, v_, evars);
	    
	    /*now print tree*/
	    for(k=0;k<strlen(newick_tree2);k++)newick_tree2[k]='\0';
	    print_rooted_tree(T, num_to_name, newick_tree2, 1);
	    printf("%s\n", newick_tree2);
	    
	    /*restore tree*/
	    del_edge(T->G, ne);
	    if(u_!=u){
	      del_node(T->G, u_);
	      evars = new_edge_aux_vars(); new_edge(T->G, lu, lv, evars);
	      evars = new_edge_aux_vars(); new_edge(T->G, nu[0], u, evars);
	      evars = new_edge_aux_vars(); new_edge(T->G, nu[1], u, evars);
	      del_edge(T->G, d1); 
	    }
	    if(v_!=v){
	      del_node(T->G, v_);
	      evars = new_edge_aux_vars(); new_edge(T->G, ru, rv, evars);
	      evars = new_edge_aux_vars(); new_edge(T->G, nv[0], v, evars);
	      evars = new_edge_aux_vars(); new_edge(T->G, nv[1], v, evars);
	      del_edge(T->G, d2);
	    }
	    evars = new_edge_aux_vars(); e = new_edge(T->G, u, v, evars);
	    
	    for(u_=T->G->first_node; u_; u_=u_->next)
	      assert(u_->degree <= 3);
	    
	    /*print for debugging*/
	    /*
	      for(k=0;k<strlen(newick_tree2);k++)newick_tree2[k]='\0';
	      print_rooted_tree(T, num_to_name, newick_tree2, 1);
	      printf("%s\n", newick_tree2);
	    */
	    /*end debug*/
	  }
	}
	return 1;
      }
      i++;
    }
  }
  return 1;
}

static void remove_2degree_butnotroot_tree(phylogenetic_tree T, const node root){
  node u, v, w, x; edge e1, e2;
  edge_aux_vars edge_vars; 
  
  assert(T);
  u = T->G->first_node;
  while(u){
    v = u->next;
    /*if(u->degree == 2 && size_int_set(((node_aux_vars)u->aux_vars)->taxaset)==0) {*/
    if(u->degree == 2 && u!=root){
      if(((node_aux_vars)u->aux_vars)->taxon_id == -1) {
        e1 = u->first_edge; e2 = u->last_edge;
        if(e1->u != u) w = e1->u; else w = e1->v;
        if(e2->u != u) x = e2->u; else x = e2->v;
        
        edge_vars = new_edge_aux_vars();
        if(e1->aux_vars && e2->aux_vars)
          edge_vars->wgt = ((edge_aux_vars)e1->aux_vars)->wgt 
            + ((edge_aux_vars)e2->aux_vars)->wgt;
        else 
          edge_vars->wgt = 0;
        
        new_edge(T->G, w, x, edge_vars);
        
        del_node(T->G, u); u = NULL;
      }
      else assert(0); /*internal 2-degree node should have -ve taxon id*/
    }
    u = v;
  }
}

static int min(int x, int y){
  if(x<y)return x;
  else return y;
}

static void permutearray(int* p, int size){
  int i, k, N, M; int* q;
  N = size; M = N;
  q = malloc(N*sizeof(int));
  for (i = 0; i < N; i++) q[i] = 0 + i;
  for (i = 0; i < M; i++){
    k=i+(int)((N-i)*((double) rand()/(double) RAND_MAX));/*uniform on i...N-1*/
    p[i] = q[k]; q[k] = q[i]; /*swap i and k terms but put i-th term into p*/
  }
  free(q);
}

/*do BFS to get edges*/
static void get_edges(list q_parent, list q_child, const phylogenetic_tree T, list l){
  
  edge f; node v; node parent; node root;
  list q_parent_; list q_child_;
  pair p;
  
  /*stopping condition*/
  if(!size_list(q_parent)) return;
  
  /*
    printf("assert..hello\n");
    printf("got %d parents and %d children\n", size_list(q_parent), size_list(q_child));
  */
  assert(size_list(q_parent) && size_list(q_child));
  q_parent_=new_list(); q_child_=new_list();
  
  while(size_list(q_parent)){
    /*HACK*/
    parent=(node) q_parent->head->data;
    q_parent->head->data=NULL; pop_list(q_parent);
    root=(node) q_child->head->data;
    q_child->head->data=NULL; pop_list(q_child);
    /*end of HACK*/
    for(f=root->first_edge; f;){
      v=opposite(root,f);
      if(v!=parent){
	if(v->degree>1){
	  append_list(q_parent_, (void *)root);
	  append_list(q_child_, (void *)v);
	}
	p=malloc(sizeof(struct pair_struct));
	p->u=f->u; p->v=f->v; append_list(l, (void*) p);
	/*append_list(l, (void*) f);*/
      }
      if(f->u==root)f=f->u_next;else if(f->v==root)f=f->v_next; else assert(0);
    }
  }
  get_edges(q_parent_, q_child_, T, l);
  assert(!size_list(q_parent_) && !size_list(q_child_));
  del_list(q_parent_); del_list(q_child_);
}
