/*
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<assert.h>
#include<ugraph_aux_vars.h>

int VERBOSE=0;

static double nbgweights(node r, edge e, int rec_level, int* c);

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  char newick_tree2[MAX_NEWICK_STR_LEN];
  char* num_to_name[MAX_LEAVES];
  /*char command[MAX_BUFFER];*/

  FILE* fp; int rec_level, i, k, edge_num, c; phylogenetic_tree T;
  edge e, f, g; node u, v, w; edge_aux_vars evars; node_aux_vars nvars;
  double total; 

  /**Initialization**/
  T = NULL;
  edge_num = -1;

  if(argc < 2){
    printf("Type roottree -help on how to run\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in the newick tree */
    if (!strcmp(argv[i], "-t")) {
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      assert(fscanf(fp, "%s", newick_tree));
      fclose(fp);

      T = new_phylogenetic_tree();
      if(!read_tree(T, num_to_name, newick_tree, 0)) {
	printf("Error: could not read tree %s\n", newick_tree);
	exit(EXIT_FAILURE);
      }
    }
    else if (!strcmp(argv[i], "-e")){
      i++;
      sscanf(argv[i], "%d", &edge_num);
      assert(edge_num >= 0);
    }
    else if (!strcmp(argv[i], "-r")){
      i++;
      sscanf(argv[i], "%d", &rec_level);
      assert(rec_level >= 1);
    }
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%d", &VERBOSE);
    }
    else if (!strcmp(argv[i], "-help")){
      printf("To do....\n");
      exit(EXIT_FAILURE);
    }
    else {
      printf("Type roottree -help on how to run\n");
      exit(EXIT_FAILURE);
    }
  }
  
  if(edge_num > T->G->n_edges) { 
    printf("Error: edge_num=%d > edges in T=%d\n", edge_num, T->G->n_edges);
  }

  i=0;
  for(e = T->G->first_edge; e; e=e->next){
    if(i==edge_num){
      /*subdivide edge and create root*/
      //printf("Edge to subdivide has weight=%0.3f\n", ((edge_aux_vars)e->aux_vars)->wgt);      
      //printf("Neighboring edges have weights:");

      c=0; total=0;
      total += nbgweights(e->u, e, rec_level, &c);
      total += nbgweights(e->v, e, rec_level, &c);
      //printf("\n");
      //printf("c = %d\n", c);
      //printf("total = %.3f\n", total);

      total/=c;
      //printf("Average of neighbors = %0.3f\n", total);

      nvars = new_node_aux_vars(); u = new_node(T->G, nvars);
      evars = new_edge_aux_vars(); evars->wgt = (((edge_aux_vars)e->aux_vars)->wgt)/2;
      f = new_edge(T->G, u, e->u, evars);
      evars = new_edge_aux_vars(); evars->wgt = (((edge_aux_vars)e->aux_vars)->wgt)/2;
      g = new_edge(T->G, u, e->v, evars);
      del_edge(T->G, e);

      set_rooted_subtree_sizes(T);
      set_rooted_subtree_edgelens(T);

      v=opposite(u,u->first_edge); w=opposite(u,u->last_edge);
      assert(v); assert(w);

      //printf("Root (num seqs): %d %d\n", ((node_aux_vars)v->aux_vars)->taxa_size, ((node_aux_vars)w->aux_vars)->taxa_size); 
      //printf("Root (num edgelens): %0.3f %0.3f\n", (((edge_aux_vars)f->aux_vars)->wgt + ((node_aux_vars)v->aux_vars)->sum_edge_lens)/(2*(((node_aux_vars)v->aux_vars)->taxa_size + 1) - 3),(((edge_aux_vars)g->aux_vars)->wgt + ((node_aux_vars)w->aux_vars)->sum_edge_lens)/(2*(((node_aux_vars)w->aux_vars)->taxa_size + 1) - 3));

      /*print tree*/
      for(k=0;k<strlen(newick_tree2);k++)newick_tree2[k]='\0';
      print_rooted_tree(T, num_to_name, newick_tree2, 1);
      printf("%s\n", newick_tree2);
      
      break;
    }
    i++;
  }
  return 1;
}

static double nbgweights(node r, edge e, int rec_level, int* c){

  double total; edge f;
  
  total=0;
  if(rec_level == 0 || r->degree==1) {
    return 0;
  }
  else {
    for(f=r->first_edge; f; ){
      if(f!=e) {
        //printf(" %0.3f", ((edge_aux_vars)f->aux_vars)->wgt);
        total+=((edge_aux_vars)f->aux_vars)->wgt + 
               nbgweights(opposite(r, f), f, rec_level-1, c);
        (*c)++;
      }
      if(f->u==r)f=f->u_next;else if(f->v==r)f=f->v_next; else assert(0);
    }
  } 
  return total;
}
