/*
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<assert.h>
#include<math.h>
#include<recdcm3.h>

int VERBOSE=0;

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];
  FILE* fp; int i; 
  phylogenetic_tree T;
  char* num_to_name[MAX_LEAVES];
  char buffer[50];
  double max_subset_size;
  
  /*initialize*/
  max_subset_size = HUGE_VAL;
  /************/

  if(argc < 2){
    printf("Type recdcm3 -help for usage\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in the newick tree*/
    if (!strcmp(argv[i], "-gt")) {
      i++;
      fp = fopen(argv[i], "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", argv[i]);
        exit(EXIT_FAILURE);
      }
      fscanf(fp, "%s", newick_tree);
      fclose(fp);
    }
    else if (!strcmp(argv[i], "-v")){
      i++;
      sscanf(argv[i], "%s", buffer);
      VERBOSE = atoi(buffer);
    }
    else if (!strcmp(argv[i], "-ms")){
      i++;
      sscanf(argv[i], "%s", buffer);
      max_subset_size = strtod(buffer, 0);
    }
    else if (!strcmp(argv[i], "-help")){
      printf("Usage: recdcm3\n\t-gt <filename containing guide_tree in newick format>\n\t(tree must NOT contain any spaces)\n\t-ms <maximum subset size>\n\t(if not specified then DCM3 decomposition is used)\n");
      exit(EXIT_FAILURE);
    }
    else {
      printf("Type recdcm3 -help for usage\n");
      exit(EXIT_FAILURE);
    }
  }
  
  /*read in tree*/
  T = new_phylogenetic_tree();
  if(!read_tree(T, num_to_name, newick_tree, 0)) {
    printf("Error: could not read tree %s\n", newick_tree);
    exit(EXIT_FAILURE);
  }

  recursive_dcm3_subsets(T, num_to_name, max_subset_size);

  for(i=0;i<T->num_leaves;i++)free(num_to_name[i]);
  del_phylogenetic_tree(T);
  return 1;
}
