/*
This file implements the recursive-dcm3 decomposition.
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<split.h>
#include<dcm3.h>
#include<math.h>
#include<assert.h>

void recursive_dcm3_subsets(phylogenetic_tree T, char** num_to_name, double max_subset_size){
  
  int i, j; phylogenetic_tree U;
  int_set subset_taxa;
  edge c_edge;
  int* subsets; int* subsets_; int* separator;
  int subset_num, separator_size;

  subsets = NULL; subsets_=NULL; separator=NULL;

  /*this array must have 1 more than number of leaves. it
    stores the index to the start and end of subsets in the subsets array*/
  subsets_ = (int*) malloc((T->num_leaves+1) * sizeof(int));

  subsets = (int*) malloc(T->num_leaves * sizeof(int));
  separator = (int*) malloc(T->num_leaves * sizeof(int));
  separator_size = 0;

  c_edge = get_centroid_edge(T);
  get_dcm3_subsets(T, c_edge, subsets, subsets_, separator, &subset_num);
  for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)separator_size++;

  for(i = 0; i < subset_num-1;i++){
    /*if subset size is greater than max_size, then recurse*/
    if(subsets_[i+1] - subsets_[i] + separator_size > max_subset_size){

      /*if subset size is at least the size of the tree 
	then we can't go any smaller*/
      if(subsets_[i+1] - subsets_[i] + separator_size >= T->num_leaves){
	/*so we print subsets*/
	for(j = subsets_[i]; j < subsets_[i+1]; j++)
	  printf("%s\t", num_to_name[subsets[j]]);
	for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)
	  printf("%s\t", num_to_name[separator[j]]);
	printf("\n");
      }

      else { /*we recurse*/
	/*convert subset into an integer-set*/
	subset_taxa = new_int_set(32);
	for(j = subsets_[i]; j < subsets_[i+1]; j++){
	  insert_int_set(subset_taxa, subsets[j]);
 	}
	/*don't forget the separator*/
	for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)
	  insert_int_set(subset_taxa, separator[j]);
	/*copy tree and induce it*/
	U = new_phylogenetic_tree(); 
	assign_tree(U, T); 
	restrict_tree(U, subset_taxa);
	/*set subtree sizes*/
	set_subtree_sizes(U);
	/*now call recursive-dcm3*/
	recursive_dcm3_subsets(U, num_to_name, max_subset_size);
	del_phylogenetic_tree(U); del_int_set(subset_taxa);
      }

    }
    else {
      /*subset size is smaller than max size so we print*/
      for(j = subsets_[i]; j < subsets_[i+1]; j++)
	printf("%s\t", num_to_name[subsets[j]]);
      for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)
        printf("%s\t", num_to_name[separator[j]]);
      printf("\n");
    }
  }

  free(subsets); free(subsets_); free(separator);
  return;
}
