#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<tree2splits.h>
#include<splits2tree.h>
#include<constants.h>
#include<phylo_tree.h>
#include<split.h>
#include<ugraph_aux_vars.h>
#include<dcm3.h>
#include<math.h>
#include<assert.h>
#include<scm_merge.h>
#include<sys/types.h>
#include<unistd.h>

int VERBOSE=0;
phylogenetic_tree recursive_dcm3(phylogenetic_tree T, char** sequences, char** num_to_name, hashtable name_to_num, double max_subset_size, char* basecmd, char* basein, char* basetree);
static void read_chars(FILE* fp, char*** num_to_names_, char*** sequences_,
		       int* ntax, int* nchars);
static phylogenetic_tree solve_subtree(char** sequences, char** num_to_name, hashtable name_to_num, int_set subset_taxa, char* basecmd, char* basein, char* basetree );

/*global*/
char _recidcm3_gt_[MAX_FILENAME];
char _pid_string_[MAX_BUFFER] ;
/********/

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];

  FILE* fp; int i, j; 
  phylogenetic_tree T, U;
  double max_subset_size;
  int num_iterations, result, mpscore;
  char** sequences;
  int ntax, nchars;
  hashtable name_to_num;
  char** num_to_name;

  char basecmd[MAX_COMMAND]; char basein[MAX_FILENAME]; 
  char basetree[MAX_FILENAME]; char globalcmd[MAX_COMMAND]; 
  /*char file_suffix[MAX_FILENAME]; */
  char output_filename[MAX_FILENAME];
  char buffer[MAX_BUFFER]; char gt_filename[MAX_FILENAME];
  char seq_filename[MAX_FILENAME]; char mpscorecmd[MAX_COMMAND];

  char* tempstr; char* cleanup_cmd;

  pid_t pid;

  /************/
  /*initialize*/
  /************/
  sequences = NULL; T = NULL; U = NULL;
  max_subset_size = HUGE_VAL;
  name_to_num = NULL; 
  num_iterations = 10;
  basein[0] = '\0'; basetree[0] = '\0';

  strcpy(_recidcm3_gt_, "__RECIDCM3__GT__");
  pid=getpid(); sprintf(_pid_string_, "%d", pid); 
  /*file_suffix[0] = '\0';*/
  /************/

  if(argc < 2){
    printf("Please specify guide tree using \"-gt <filename containing guide_tree>\"\n");
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in sequences*/
    if (!strcmp(argv[i], "-s")){
      i++; sscanf(argv[i], "%s", seq_filename);
      assert(seq_filename[0] != '-');
      fp = fopen(seq_filename, "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", seq_filename);
        exit(EXIT_FAILURE);
      }
      read_chars(fp, &num_to_name, &sequences, &ntax, &nchars);
      fclose(fp);

      assert(num_to_name); assert(sequences);
      name_to_num = new_hashtable(ntax);
      for(j=0;j<ntax;j++){
	insert_hashtable(name_to_num, num_to_name[j], j);
      }
    }
    /*read in output filename*/
    else if (!strcmp(argv[i], "-o")){
      /*output_filename = (char*) malloc(MAX_FILENAME * sizeof(char));*/
      /*assert(output_filename);*/
      i++; sscanf(argv[i], "%s", output_filename);
      assert(output_filename[0] != '-');
    }
    else if (!strcmp(argv[i], "-gt")) {
      i++; sscanf(argv[i], "%s", gt_filename);
      assert(gt_filename[0] != '-');
      fp = fopen(gt_filename, "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", gt_filename);
        exit(EXIT_FAILURE);
      }
      fscanf(fp, "%s", newick_tree);
      fclose(fp);
      
      /*read in guide tree*/
      assert(name_to_num);
      T = new_phylogenetic_tree();
      if(!read_tree(T, num_to_name, newick_tree, name_to_num)) {
	printf("Error: could not read tree %s\n", newick_tree);
	exit(EXIT_FAILURE);
      }
    }
    else if (!strcmp(argv[i], "-v")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      VERBOSE = atoi(buffer);
    }
    else if (!strcmp(argv[i], "-ms")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      max_subset_size = strtod(buffer, 0);
    }
    else if (!strcmp(argv[i], "-it")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      num_iterations = atoi(buffer);
    }
    else if (!strcmp(argv[i], "-bin")){
      i++; sscanf(argv[i], "%s", basein);
      assert(basein[0] != '-');
    }
    else if (!strcmp(argv[i], "-bout")){
      i++; sscanf(argv[i], "%s", basetree);
      assert(basetree[0] != '-');
    }
    else if (!strcmp(argv[i], "-bcmd")){
      i++; assert(!strcmp(argv[i], "[")); i++;
      while(i < argc && strcmp(argv[i], "]")){
        strcat(basecmd, argv[i]); i++;
        strcat(basecmd, " "); 
      }
      if(i==argc) { printf("Error: no \']\' found\n"); exit(EXIT_FAILURE); }
    }
    else if (!strcmp(argv[i], "-gcmd")){
      i++; assert(!strcmp(argv[i], "[")); i++;
      while(i < argc && strcmp(argv[i], "]")){
        strcat(globalcmd, argv[i]); i++;
        strcat(globalcmd, " "); 
      }
      if(i==argc) { printf("Error: no \']\' found\n"); exit(EXIT_FAILURE); }
    }
    else {
      printf("Please specify guide tree using \"-gt <filename containing guide_tree>\"\n");
      exit(EXIT_FAILURE);
    }
  }

  /*setting file names properly*/
  if(!basein[0])strcpy(basein,"basein"); 
  if(!basetree[0])strcpy(basetree,"basetree");
  strcat(_recidcm3_gt_, _pid_string_); strcat(basein, _pid_string_);
  strcat(basetree, _pid_string_); 

  /*creating base and global command strings*/
  strcat(basecmd, " "); strcat(basecmd, basein); strcat(basecmd, " "); 
  strcat(basecmd, _recidcm3_gt_); strcat(basecmd, " ");
  strcat(basecmd, basetree); strcat(basecmd," "); strcat(basecmd,_pid_string_);
  strcat(mpscorecmd, "perl paup_mpscore.pl "); strcat(mpscorecmd, seq_filename); 
  strcat(mpscorecmd, " "); strcat(mpscorecmd, _recidcm3_gt_); 
  strcat(mpscorecmd, " "); strcat(mpscorecmd, _pid_string_);
  strcat(globalcmd, " "); strcat(globalcmd, seq_filename); strcat(basecmd, " "); 
  strcat(globalcmd, _recidcm3_gt_); strcat(globalcmd, " ");
  strcat(globalcmd,_pid_string_);
  
  printf("globalcmd=%s, basecmd=%s, basein=%s, basetree=%s\n", globalcmd, basecmd, basein, basetree);  
  printf("num_iterations=%d, maxsize=%.2f,\n", num_iterations, max_subset_size);

  assert(T && sequences);

  /*keeping a copy of current guide-tree in file called __RECIDCM3__GT__*/
  fp = fopen(_recidcm3_gt_, "w");
  if(!fp) { 
    printf("Error: could not open filename \"%s\"\n", _recidcm3_gt_); 
    exit(EXIT_FAILURE);
  } 
  fprintf(fp, "%s", newick_tree); fclose(fp);
  
  for(j=0; j<num_iterations; j++){
    U = recursive_dcm3(T, sequences, num_to_name, name_to_num, max_subset_size, basecmd, basein, basetree);
    printf("U has %d leaves\n", U->num_leaves);
    assert(U);
    for(i=0;i<strlen(newick_tree);i++) newick_tree[i]='\0';
    print_tree(U, num_to_name, newick_tree, 1);
    /*keeping a copy of current guide-tree in file called __RECIDCM3__GT__*/
    fp = fopen(_recidcm3_gt_, "w");
    if(!fp) {
      printf("Error: could not open filename \"%s\"\n", _recidcm3_gt_);
      exit(EXIT_FAILURE);
    }
    fprintf(fp, "%s", newick_tree); fclose(fp);

    /*get MP score*/
    printf("mp score cmd=%s\n", mpscorecmd);
    result = system(mpscorecmd);
    printf("result=%d\n", result);
    if(result == -1) {
      printf("Error: could not execute system command \"%s\"\n", mpscorecmd);
      exit(EXIT_FAILURE);
    }

    tempstr=(char*)malloc(MAX_FILENAME * sizeof(char));
    tempstr[0]='\0'; strcat(tempstr, "paup_mpscore"); strcat(tempstr,_pid_string_);
    printf("opening %s\n", tempstr);
    fp=fopen(tempstr, "r");
    if(!fp) {
      printf("Error: could not open filename \"%s\"\n", tempstr);
      exit(EXIT_FAILURE);
    }
    fscanf(fp, "%s", buffer); fscanf(fp, "%s", buffer);
    fscanf(fp, "%s", buffer); fscanf(fp, "%s", buffer);
    printf("read %s\n", buffer); fclose(fp);
    sscanf(buffer, "%d", &mpscore);
    printf("mp score=%d\n", mpscore);
    
    /*do local search*/

    /*get MP score*/
  }
  
  del_phylogenetic_tree(T);
  del_phylogenetic_tree(U);
  for(i=0;i<ntax;i++)free(sequences[i]); free(sequences);
  del_hashtable(name_to_num);

  cleanup_cmd=(char*)malloc(MAX_COMMAND * sizeof(char));
  cleanup_cmd[0]='\0'; strcat(cleanup_cmd,"/bin/rm ");
  strcat(cleanup_cmd, _recidcm3_gt_);
  result = system(cleanup_cmd);
  if(result == -1) {
    printf("Error: could not execute system command \"%s\"\n", cleanup_cmd);
    exit(EXIT_FAILURE);
  }
  
  return 1;
}

phylogenetic_tree recursive_dcm3(phylogenetic_tree T, char** sequences, char** num_to_name, hashtable name_to_num, double max_subset_size, char* basecmd, char* basein, char* basetree){
  
  int i, j; phylogenetic_tree U;
  int_set subset_taxa;
  edge c_edge;
  int* subsets; int* subsets_; int* separator;
  int subset_num, separator_size;
  phylogenetic_tree* tree_array;
  phylogenetic_tree recidcm3_tree;

  recidcm3_tree = NULL;
  subsets = NULL; subsets_=NULL; separator=NULL;

  subsets_ = (int*) malloc((T->num_leaves+1) * sizeof(int));
  subsets = (int*) malloc(T->num_leaves * sizeof(int));
  separator = (int*) malloc(T->num_leaves * sizeof(int));
  separator_size = 0;

  c_edge = get_centroid_edge(T);
  get_dcm3_subsets(T, c_edge, subsets, subsets_, separator, &subset_num);
  
  tree_array = (phylogenetic_tree*) 
    malloc((subset_num-1) * sizeof(struct phylogenetic_tree_struct));
  
  for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)separator_size++;

  for(i = 0; i < subset_num-1;i++){
    /*if subset size is greater than max_size, then recurse*/
    if(subsets_[i+1] - subsets_[i] + separator_size > max_subset_size){

      /*if subset size is the same size as the tree 
	then we can't go any smaller*/
      if(subsets_[i+1] - subsets_[i] + separator_size >= T->num_leaves){
	/*so we print subsets*/
	subset_taxa = new_int_set(32);
	for(j = subsets_[i]; j < subsets_[i+1]; j++){
	  /*printf("%s\t", num_to_name[subsets[j]]);*/
	  insert_int_set(subset_taxa, subsets[j]);
	}
	for(j = 0; separator[j] != -1 && j < T->num_leaves;j++){
	  /*printf("%s\t", num_to_name[separator[j]]);*/
	  insert_int_set(subset_taxa, separator[j]);
	}
	/*printf("\n");*/

	/*compute subtree*/
        /*printf("solving subproblem %d\n", i);*/
        tree_array[i] = solve_subtree(sequences, num_to_name, name_to_num, subset_taxa, basecmd, basein, basetree);
        /*printf("solved subproblem %d\n", i);*/
      }

      else { /*we recurse*/
	/*convert subset into an integer-set*/

	subset_taxa = new_int_set(32);
	for(j = subsets_[i]; j < subsets_[i+1]; j++)
	  insert_int_set(subset_taxa, subsets[j]);
	/*don't forget the separator*/
	for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)
	  insert_int_set(subset_taxa, separator[j]);
	/*copy tree and induce it*/
	U = new_phylogenetic_tree(); 
	assign_tree(U, T); 
	induce_tree(U, subset_taxa);
	/*set subtree sizes*/
	set_subtree_sizes(U);
	/*now call recursive-dcm3*/
	tree_array[i] = recursive_dcm3(U, sequences, num_to_name, name_to_num,  max_subset_size, basecmd, basein, basetree);
	del_phylogenetic_tree(U); del_int_set(subset_taxa);
      }

    }
    else {
      /*print subsets*/
      subset_taxa = new_int_set(32);
      for(j = subsets_[i]; j < subsets_[i+1]; j++){
	/*printf("%s\t", num_to_name[subsets[j]]);*/
	insert_int_set(subset_taxa, subsets[j]);	
      }
      for(j = 0; separator[j] != -1 && j < T->num_leaves;j++){
        /*printf("%s\t", num_to_name[separator[j]]);*/
	insert_int_set(subset_taxa, separator[j]);	
      }
      /*printf("\n");*/
      
      /*compute subtree*/
      /*printf("solving subproblem %d\n", i);*/
      tree_array[i] = solve_subtree(sequences, num_to_name, name_to_num, subset_taxa, basecmd, basein, basetree);
      /*printf("solved subproblem %d\n", i);*/
    }
  }

  /*printf("computed subtrees\n");*/

  free(subsets); free(subsets_); free(separator);

  /*merge trees*/
  recidcm3_tree = scm_merge(tree_array, subset_num - 1);

  /*free tree_array*/
  for(i=0;i<subset_num-1;i++) del_phylogenetic_tree(tree_array[i]);
  
  /*random refine recidcm3 tree*/
  random_refine(recidcm3_tree);

  printf("returning\n");
  return recidcm3_tree;
}

static phylogenetic_tree solve_subtree(char** sequences, char** num_to_name, hashtable name_to_num, int_set subset_taxa, char* basecmd, char* basein, char* basetree ){
  
  FILE* fp; int result;
  char newick_tree[MAX_NEWICK_STR_LEN];
  /*char cleanup_cmd[MAX_COMMAND];
    char cleanup_cmd2[MAX_COMMAND];*/
  char* cleanup_cmd;
  char* cleanup_cmd2;
  phylogenetic_tree T; int_set tree_taxa;
  node v; int i; int* entries;

  assert(size_int_set(subset_taxa) > 3);

  /*prepare cleanup commands --- can pass these later as parameters*/
  cleanup_cmd=(char*)malloc(MAX_COMMAND * sizeof(char));
  cleanup_cmd2=(char*)malloc(MAX_COMMAND * sizeof(char));
  cleanup_cmd[0]='\0'; cleanup_cmd2[0]='\0';
  strcat(cleanup_cmd, "/bin/rm "); strcat(cleanup_cmd, basein);
  strcat(cleanup_cmd2, "/bin/rm "); strcat(cleanup_cmd2, basetree);
  /*printf("cleanup=%s\n", cleanup_cmd);
    printf("cleanup2=%s\n", cleanup_cmd2);*/
  
  /*print sequences to file*/
  fp = fopen(basein, "w");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", basein);
    exit(EXIT_FAILURE);
  }

  fprintf(fp, "%d %d\n", size_int_set(subset_taxa), strlen(sequences[0]));
  entries=(int*)malloc(size_int_set(subset_taxa) * sizeof(int));
  get_entries_int_set(subset_taxa, entries);
  for(i=0;i<size_int_set(subset_taxa);i++){
    assert(num_to_name[entries[i]]);
    fprintf(fp, "%-10s%s\n", num_to_name[entries[i]], sequences[entries[i]]);
  }
  fclose(fp);

  /*printf("base cmd=%s\n", basecmd);*/
  result = system(basecmd);
  printf("result=%d\n", result);
  if(result == -1) {
    printf("Error: could not execute system command \"%s\"\n", basecmd);
    exit(EXIT_FAILURE);
  }
  
  /*remove basechars*/
  result = system(cleanup_cmd);
  if(result == -1) {
    printf("Error: could not execute system command \"%s\"\n", cleanup_cmd);
    exit(EXIT_FAILURE);
  }
  
  /*read in base tree*/
  fp = fopen(basetree, "r");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", basetree);
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%s", newick_tree);
  fclose(fp);
  
  /*remove basetree*/
  result = system(cleanup_cmd2);
  if(result == -1) {
    printf("Error: could not execute system command \"%s\"\n", cleanup_cmd2);
    exit(EXIT_FAILURE);
  }

  T = new_phylogenetic_tree();
  if(!read_tree(T, num_to_name, newick_tree, name_to_num)) {
    printf("Error: could not read tree %s\n", newick_tree);
    exit(EXIT_FAILURE);
  }
  
  /*check to make sure taxa of tree read and the subset_taxa match*/
  tree_taxa = new_int_set(32);
  for(v = T->G->first_node; v; v = v->next)
    if(v->degree==1) {
      assert(((node_aux_vars)v->aux_vars)->taxon_id != -1);
      insert_int_set(tree_taxa, ((node_aux_vars)v->aux_vars)->taxon_id);
    }
  /*print_int_set(tree_taxa); printf("\n");
  print_int_set(subset_taxa); printf("\n");*/
  assert(equal_int_set(tree_taxa, subset_taxa));

  free(cleanup_cmd); free(cleanup_cmd2);

  return T;
}
  
static void read_chars(FILE* fp, char*** num_to_names_, char*** sequences_,
		int* ntax, int* nchars){
  
  int i; char** sequences; char** num_to_names;

  *ntax=0; *nchars=0;
  fscanf(fp, "%d", ntax);
  fscanf(fp, "%d", nchars);
  if(*ntax<=0 || *nchars<=0){
    printf("Error: number of taxa=%d and number of sites=%d must be positive", *ntax, *nchars);
    exit(EXIT_FAILURE);
  }
  
  sequences=(char**) malloc(*ntax * sizeof(char*));
  num_to_names=(char**) malloc(*ntax * sizeof(char*));
  for(i=0;i<*ntax;i++){
    sequences[i] = (char*) malloc((*nchars+1) * sizeof(char));
    num_to_names[i] = (char*) malloc(MAX_NAME_LEN * sizeof(char));
    fscanf(fp, "%s", num_to_names[i]);
    fscanf(fp, "%s", sequences[i]);
  }

  /*
  for(i=0;i<*ntax;i++)
    printf("%s\t%s\n", num_to_names[i], sequences[i]);
  */

  *sequences_ = sequences;
  *num_to_names_ = num_to_names;
}

