#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<tree2splits.h>
#include<splits2tree.h>
#include<constants.h>
#include<phylo_tree.h>
#include<split.h>
#include<ugraph_aux_vars.h>
#include<dcm3.h>
#include<math.h>
#include<assert.h>
#include<scm_merge.h>
#include<sys/types.h>
#include<unistd.h>
#include<signal.h>

int VERBOSE=0;
phylogenetic_tree recursive_dcm3(phylogenetic_tree T, char** sequences, char** num_to_name, hashtable name_to_num, double max_subset_size, char* basecmd, char* basein, char* basetree);

/*static functions*/
static void read_chars(FILE* fp, char*** num_to_names_, char*** sequences_,
		       int* ntax, int* nchars);
static phylogenetic_tree solve_subtree(char** sequences, char** num_to_name, hashtable name_to_num, int_set subset_taxa, char* basecmd, char* basein, char* basetree );
static phylogenetic_tree global_search(char* globalcmd, char** num_to_name, hashtable name_to_num, int_set full_taxa, int* treescore);
static int get_mpscore(char* mpscore_cmd);
static void cleanup(int dummy);

/*************/

/*global*/
char _recidcm3_gt_[MAX_FILENAME];
/*char _recidcm3_gt_temp_[MAX_FILENAME];*/
char _pid_string_[MAX_BUFFER] ;
char basein[MAX_FILENAME]; char basetree[MAX_FILENAME]; 
char scorefile[MAX_FILENAME];
/********/

int main(int argc, char* argv[]){

  char newick_tree[MAX_NEWICK_STR_LEN];

  FILE* fp; phylogenetic_tree T, U;
  double max_subset_size;
  int num_iterations, result, mpscore, ntax, nchars, i, j;

  char** sequences;  char** num_to_name;
  hashtable name_to_num;

  char basecmd[MAX_COMMAND]; char globalcmd[MAX_COMMAND]; 
  char output_filename[MAX_FILENAME];
  char buffer[MAX_BUFFER]; char gt_filename[MAX_FILENAME];
  char seq_filename[MAX_FILENAME]; char mpscorecmd[MAX_COMMAND];

  int_set full_taxa;

  pid_t pid;

  /************/
  /*initialize*/
  /************/
  signal(SIGTERM, (void (*)(int)) cleanup);
  signal(SIGABRT, (void (*)(int)) cleanup);
  sequences = NULL; T = NULL; U = NULL;
  max_subset_size = HUGE_VAL;
  name_to_num = NULL; 
  num_iterations = 10;
  basein[0] = '\0'; basetree[0] = '\0';

  strcpy(_recidcm3_gt_, "RECIDCM3GT");
/*  strcpy(_recidcm3_gt_temp_, "RECIDCM3GTTEMP");*/
  strcpy(basein,"BASECHARS"); 
  strcpy(basetree,"BASETREE");
  strcpy(scorefile,"TREESCORE");

  pid=getpid(); sprintf(_pid_string_, "%d", pid); 
  strcat(_recidcm3_gt_, _pid_string_); strcat(basein, _pid_string_);
  strcat(basetree, _pid_string_); strcat(scorefile, _pid_string_); 
  /************/

  if(argc < 2){
    printf("Please specify guide tree using \"-gt <filename containing guide_tree>\"\n");
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  for(i = 1; i < argc; i++) {
    /*read in sequences*/
    if (!strcmp(argv[i], "-s")){
      i++; sscanf(argv[i], "%s", seq_filename);
      assert(seq_filename[0] != '-');
      fp = fopen(seq_filename, "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", seq_filename);
        cleanup(1);
        exit(EXIT_FAILURE);
      }
      read_chars(fp, &num_to_name, &sequences, &ntax, &nchars);
      fclose(fp);
      assert(num_to_name); assert(sequences);
      name_to_num = new_hashtable(ntax);
      full_taxa = new_int_set(32);
      for(j=0;j<ntax;j++){
	insert_int_set(full_taxa, j);
	insert_hashtable(name_to_num, num_to_name[j], j);
      }
    }
    /*read in output filename*/
    else if (!strcmp(argv[i], "-o")){
      i++; sscanf(argv[i], "%s", output_filename);
      assert(output_filename[0] != '-');
    }
    else if (!strcmp(argv[i], "-gt")) {
      i++; sscanf(argv[i], "%s", gt_filename);
      assert(gt_filename[0] != '-');
      fp = fopen(gt_filename, "r");
      if(!fp) {
        printf("Error: could not open filename \"%s\"\n", gt_filename);
        cleanup(1);
        exit(EXIT_FAILURE);
      }
      fscanf(fp, "%s", newick_tree);
      fclose(fp);
      
      /*read in guide tree*/
      assert(name_to_num);
      T = new_phylogenetic_tree();
      if(!read_tree(T, num_to_name, newick_tree, name_to_num)) {
	printf("Error: could not read tree %s\n", newick_tree);
        cleanup(1);
	exit(EXIT_FAILURE);
      }
    }
    else if (!strcmp(argv[i], "-v")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      VERBOSE = atoi(buffer);
    }
    else if (!strcmp(argv[i], "-ms")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      max_subset_size = strtod(buffer, 0);
    }
    else if (!strcmp(argv[i], "-it")){
      i++; sscanf(argv[i], "%s", buffer);
      assert(buffer[0] != '-');
      num_iterations = atoi(buffer);
    }
    else if (!strcmp(argv[i], "-bcmd")){
      i++; assert(!strcmp(argv[i], "[")); i++;
      while(i < argc && strcmp(argv[i], "]")){
        strcat(basecmd, argv[i]); i++;
        strcat(basecmd, " "); 
      }
      if(i==argc) { printf("Error: no \']\' found\n"); cleanup(1); exit(EXIT_FAILURE); }
    }
    else if (!strcmp(argv[i], "-gcmd")){
      i++; assert(!strcmp(argv[i], "[")); i++;
      while(i < argc && strcmp(argv[i], "]")){
        strcat(globalcmd, argv[i]); i++;
        strcat(globalcmd, " "); 
      }
      if(i==argc) { printf("Error: no \']\' found\n"); cleanup(1); exit(EXIT_FAILURE); }
    }
    else {
      printf("Please specify guide tree using \"-gt <filename containing guide_tree>\"\n");
      cleanup(1);
      exit(EXIT_FAILURE);
    }
  }

  /*creating base, mpscore, and global command strings*/
  strcat(basecmd, " "); strcat(basecmd, basein); strcat(basecmd, " "); 
  strcat(basecmd, _recidcm3_gt_); strcat(basecmd, " ");
  strcat(basecmd, basetree); strcat(basecmd," "); strcat(basecmd,_pid_string_);

  strcat(mpscorecmd, "perl paup_mpscore.pl "); strcat(mpscorecmd, seq_filename); 
  strcat(mpscorecmd, " "); strcat(mpscorecmd, _recidcm3_gt_); 
  strcat(mpscorecmd, " "); strcat(mpscorecmd, _pid_string_);

  strcat(globalcmd, " "); strcat(globalcmd, seq_filename); strcat(globalcmd, " "); 
  strcat(globalcmd, _recidcm3_gt_); strcat(globalcmd, " ");
  strcat(globalcmd, basetree); strcat(globalcmd," ");  
  strcat(globalcmd, scorefile); strcat(globalcmd, " ");
  strcat(globalcmd,_pid_string_);
  
  printf("globalcmd=%s\n", globalcmd);
  printf("basecmd=%s\n", basecmd); 
  printf("num_iterations=%d, maxsize=%.2f,\n", num_iterations, max_subset_size);

  assert(T && sequences);

  /*keeping a copy of current guide-tree in file called RECIDCM3GT*/
  fp = fopen(_recidcm3_gt_, "w");
  if(!fp) { 
    printf("Error: could not open filename \"%s\"\n", _recidcm3_gt_); 
    cleanup(1);
    exit(EXIT_FAILURE);
  } 
  fprintf(fp, "%s", newick_tree); fclose(fp);
  
  for(j=0; j<num_iterations; j++){
    printf("doing iteration %d...\n", j+1);
    U = recursive_dcm3(T, sequences, num_to_name, name_to_num, max_subset_size, basecmd, basein, basetree);
    assert(U);

    /*keeping a copy of current guide-tree in file called __RECIDCM3__GT__*/
    for(i=0;i<strlen(newick_tree);i++) newick_tree[i]='\0';
    print_tree(U, num_to_name, newick_tree, 0);
    fp = fopen(_recidcm3_gt_, "w");
    if(!fp) {
      printf("Error: could not open filename \"%s\"\n", _recidcm3_gt_);
      cleanup(1);
      exit(EXIT_FAILURE);
    }
    fprintf(fp, "%s", newick_tree); fclose(fp);
    /**********/

    del_phylogenetic_tree(T);
    del_phylogenetic_tree(U);

    /*get_mpscore gets mpscore of tree in _recidcm3_gt_*/
    /*mpscore = get_mpscore(mpscorecmd);
    printf("mpscore=%d\n", mpscore);*/

    /*do global search and copy result into _recidcm3_gt_*/
    T = global_search(globalcmd, num_to_name, name_to_num, full_taxa, &mpscore);

    /*
    for(i=0;i<strlen(newick_tree);i++) newick_tree[i]='\0';
    print_tree(T, num_to_name, newick_tree, 0);
    fp = fopen(_recidcm3_gt_, "w");
    if(!fp) {
      printf("Error: could not open filename \"%s\"\n", _recidcm3_gt_);
      cleanup(1);
      exit(EXIT_FAILURE);
    }
    fprintf(fp, "%s", newick_tree); fclose(fp);
    */
    /*get MP score*/
    /*mpscore = get_mpscore(mpscorecmd);*/

    printf("mpscore=%d\n", mpscore);
    
    /*read in base tree*/
    /*
    for(i=0;i<strlen(newick_tree);i++) newick_tree[i]='\0';
    fp = fopen(_recidcm3_gt_, "r");
    if(!fp) {
      printf("Error: could not open filename \"%s\"\n", _recidcm3_gt_);
      cleanup(1);
      exit(EXIT_FAILURE);
    }
    fscanf(fp, "%s", newick_tree);
    fclose(fp);
    
    del_phylogenetic_tree(T);
    T = new_phylogenetic_tree();
    if(!read_tree(T, num_to_name, newick_tree, name_to_num)) {
      printf("Error: could not read tree %s\n", newick_tree);
      cleanup(1);
      exit(EXIT_FAILURE);
    }
    assert(T); assert(T->num_leaves>0);
    */
  }
  
  cleanup(1);

  /*free memory*/
  del_phylogenetic_tree(T);
  for(i=0;i<ntax;i++)free(sequences[i]); free(sequences);
  for(i=0;i<ntax;i++)free(num_to_name[i]); free(num_to_name);
  del_hashtable(name_to_num);
  del_int_set(full_taxa);
  /*******/
  
  return 1;
}

static void cleanup(int dummy){

  char* cleanup_cmd; 
  int result;

  /*remove files*/
  cleanup_cmd=(char*)malloc(MAX_COMMAND * sizeof(char));
  cleanup_cmd[0]='\0'; strcat(cleanup_cmd,"/bin/rm ");
  strcat(cleanup_cmd, _recidcm3_gt_); strcat(cleanup_cmd, " ");
  strcat(cleanup_cmd, basein); strcat(cleanup_cmd, " ");
  strcat(cleanup_cmd, basetree); strcat(cleanup_cmd, " ");
  strcat(cleanup_cmd, "paup_mpscore"); strcat(cleanup_cmd, _pid_string_);
  strcat(cleanup_cmd, " "); strcat(cleanup_cmd, "temptree"); 
  strcat(cleanup_cmd, _pid_string_);
  result = system(cleanup_cmd);
  if(result == -1) {
    printf("Error: could not execute system command \"%s\"\n", cleanup_cmd);
    exit(EXIT_FAILURE);
  }
  free(cleanup_cmd);
  /*********/
}

static phylogenetic_tree global_search(char* globalcmd, char** num_to_name, hashtable name_to_num, int_set full_taxa, int* treescore){
  
  int result;
  char newick_tree[MAX_NEWICK_STR_LEN];
  phylogenetic_tree T; FILE* fp;
  int_set tree_taxa; node v;
  char* mv_cmd;

  result = system(globalcmd);
  if(result == -1) {
    printf("Error: could not execute system command \"%s\"\n", globalcmd);
    cleanup(1);
    exit(EXIT_FAILURE);
  }

  /*read in tree score*/
  fp = fopen(scorefile, "r");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", scorefile);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%d", treescore);
  fclose(fp);

  /*copy tree with branch lengths into _recidcm3_gt_*/
  mv_cmd=(char*)malloc(MAX_COMMAND * sizeof(char));mv_cmd[0]='\0';
  strcat(mv_cmd, "mv "); strcat(mv_cmd, basetree);
  strcat(mv_cmd, " "); strcat(mv_cmd, _recidcm3_gt_);
  result = system(mv_cmd);
  if(result == -1) {
    printf("Error: could not execute system command \"%s\"\n", mv_cmd);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  free(mv_cmd);

  /*read in new tree*/
  fp = fopen(_recidcm3_gt_, "r");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", _recidcm3_gt_);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%s", newick_tree);
  fclose(fp);
  
  T = new_phylogenetic_tree();
  if(!read_tree(T, num_to_name, newick_tree, name_to_num)) {
    printf("Error: could not read tree %s\n", newick_tree);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  /*check to make sure taxa of tree read and the subset_taxa match*/
  tree_taxa = new_int_set(32);
  for(v = T->G->first_node; v; v = v->next)
    if(v->degree==1) {
      assert(((node_aux_vars)v->aux_vars)->taxon_id != -1);
      insert_int_set(tree_taxa, ((node_aux_vars)v->aux_vars)->taxon_id);
    }
  assert(equal_int_set(tree_taxa, full_taxa));
  del_int_set(tree_taxa);

  return T;
}

static int get_mpscore(char* mpscorecmd){
  
  int mpscore, result; char* tempstr; FILE* fp;
  char buffer[MAX_BUFFER];

  result = system(mpscorecmd);
  if(result == -1) {
    printf("Error: could not execute system command \"%s\"\n", mpscorecmd);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  tempstr=(char*)malloc(MAX_FILENAME * sizeof(char));
  tempstr[0]='\0'; strcat(tempstr, "paup_mpscore"); strcat(tempstr,_pid_string_);
  fp=fopen(tempstr, "r");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", tempstr);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  free(tempstr);

  fscanf(fp, "%s", buffer); fscanf(fp, "%s", buffer);
  fscanf(fp, "%s", buffer); fscanf(fp, "%s", buffer);
  sscanf(buffer, "%d", &mpscore);
  return mpscore;
}

phylogenetic_tree recursive_dcm3(phylogenetic_tree T, char** sequences, char** num_to_name, hashtable name_to_num, double max_subset_size, char* basecmd, char* basein, char* basetree){
  
  int i, j; phylogenetic_tree U;
  int_set subset_taxa;
  edge c_edge;
  int* subsets; int* subsets_; int* separator;
  int subset_num, separator_size;
  phylogenetic_tree* tree_array;
  phylogenetic_tree recidcm3_tree;

  recidcm3_tree = NULL;
  subsets = NULL; subsets_=NULL; separator=NULL;

  subsets_ = (int*) malloc((T->num_leaves+1) * sizeof(int));
  subsets = (int*) malloc(T->num_leaves * sizeof(int));
  separator = (int*) malloc(T->num_leaves * sizeof(int));
  separator_size = 0;

  c_edge = get_centroid_edge(T);
  get_dcm3_subsets(T, c_edge, subsets, subsets_, separator, &subset_num);
  
  tree_array = (phylogenetic_tree*) 
    malloc((subset_num-1) * sizeof(struct phylogenetic_tree_struct));
  
  for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)separator_size++;

  for(i = 0; i < subset_num-1;i++){
    /*if subset size is greater than max_size, then recurse*/
    if(subsets_[i+1] - subsets_[i] + separator_size > max_subset_size){

      /*if subset size is the same size as the tree 
	then we can't go any smaller*/
      if(subsets_[i+1] - subsets_[i] + separator_size >= T->num_leaves){
	/*so we print subsets*/
	subset_taxa = new_int_set(32);
	for(j = subsets_[i]; j < subsets_[i+1]; j++){
	  /*printf("%s\t", num_to_name[subsets[j]]);*/
	  insert_int_set(subset_taxa, subsets[j]);
	}
	for(j = 0; separator[j] != -1 && j < T->num_leaves;j++){
	  /*printf("%s\t", num_to_name[separator[j]]);*/
	  insert_int_set(subset_taxa, separator[j]);
	}
	/*printf("\n");*/

	/*compute subtree*/
        /*printf("solving subproblem %d\n", i);*/
        tree_array[i] = solve_subtree(sequences, num_to_name, name_to_num, subset_taxa, basecmd, basein, basetree);
        /*printf("solved subproblem %d\n", i);*/
	del_int_set(subset_taxa);
      }

      else { /*we recurse*/
	/*convert subset into an integer-set*/

	subset_taxa = new_int_set(32);
	for(j = subsets_[i]; j < subsets_[i+1]; j++)
	  insert_int_set(subset_taxa, subsets[j]);
	/*don't forget the separator*/
	for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)
	  insert_int_set(subset_taxa, separator[j]);
	/*copy tree and induce it*/
	U = new_phylogenetic_tree(); 
	assign_tree(U, T); 
	induce_tree(U, subset_taxa);
	/*set subtree sizes*/
	set_subtree_sizes(U);
	/*now call recursive-dcm3*/
	tree_array[i] = recursive_dcm3(U, sequences, num_to_name, name_to_num,  max_subset_size, basecmd, basein, basetree);
	del_phylogenetic_tree(U); del_int_set(subset_taxa);
      }

    }
    else {
      /*print subsets*/
      subset_taxa = new_int_set(32);
      for(j = subsets_[i]; j < subsets_[i+1]; j++){
	/*printf("%s\t", num_to_name[subsets[j]]);*/
	insert_int_set(subset_taxa, subsets[j]);	
      }
      for(j = 0; separator[j] != -1 && j < T->num_leaves;j++){
        /*printf("%s\t", num_to_name[separator[j]]);*/
	insert_int_set(subset_taxa, separator[j]);	
      }
      /*printf("\n");*/
      
      /*compute subtree*/
      /*printf("solving subproblem %d\n", i);*/
      tree_array[i] = solve_subtree(sequences, num_to_name, name_to_num, subset_taxa, basecmd, basein, basetree);
      /*printf("solved subproblem %d\n", i);*/
      del_int_set(subset_taxa);
    }
  }

  /*printf("computed subtrees\n");*/

  free(subsets); free(subsets_); free(separator);

  /*merge trees*/
  recidcm3_tree = scm_merge(tree_array, subset_num - 1);

  /*free tree_array*/
  for(i=0;i<subset_num-1;i++) del_phylogenetic_tree(tree_array[i]);
  free(tree_array);

  /*random refine recidcm3 tree*/
  random_refine(recidcm3_tree);

  return recidcm3_tree;
}

static phylogenetic_tree solve_subtree(char** sequences, char** num_to_name, hashtable name_to_num, int_set subset_taxa, char* basecmd, char* basein, char* basetree ){
  
  FILE* fp; int result;
  char newick_tree[MAX_NEWICK_STR_LEN];
  phylogenetic_tree T; int_set tree_taxa;
  node v; int i; int* entries;

  assert(size_int_set(subset_taxa) > 3);

  /*print sequences to file*/
  fp = fopen(basein, "w");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", basein);
    cleanup(1);
    exit(EXIT_FAILURE);
  }

  fprintf(fp, "%d %d\n", size_int_set(subset_taxa), strlen(sequences[0]));
  entries=(int*)malloc(size_int_set(subset_taxa) * sizeof(int));
  get_entries_int_set(subset_taxa, entries);
  for(i=0;i<size_int_set(subset_taxa);i++){
    assert(num_to_name[entries[i]]);
    fprintf(fp, "%-10s%s\n", num_to_name[entries[i]], sequences[entries[i]]);
  }
  fclose(fp); free(entries);

  /*printf("base cmd=%s\n", basecmd);*/
  result = system(basecmd);
  if(result == -1) {
    printf("Error: could not execute system command \"%s\"\n", basecmd);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  /*read in base tree*/
  fp = fopen(basetree, "r");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", basetree);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%s", newick_tree);
  fclose(fp);
  
  T = new_phylogenetic_tree();
  if(!read_tree(T, num_to_name, newick_tree, name_to_num)) {
    printf("Error: could not read tree %s\n", newick_tree);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  /*check to make sure taxa of tree read and the subset_taxa match*/
  tree_taxa = new_int_set(32);
  for(v = T->G->first_node; v; v = v->next)
    if(v->degree==1) {
      assert(((node_aux_vars)v->aux_vars)->taxon_id != -1);
      insert_int_set(tree_taxa, ((node_aux_vars)v->aux_vars)->taxon_id);
    }
  /*print_int_set(tree_taxa); printf("\n");
  print_int_set(subset_taxa); printf("\n");*/
  assert(equal_int_set(tree_taxa, subset_taxa));
  del_int_set(tree_taxa);

  return T;
}
  
static void read_chars(FILE* fp, char*** num_to_names_, char*** sequences_,
		int* ntax, int* nchars){
  
  int i; char** sequences; char** num_to_names;

  *ntax=0; *nchars=0;
  fscanf(fp, "%d", ntax);
  fscanf(fp, "%d", nchars);
  if(*ntax<=0 || *nchars<=0){
    printf("Error: number of taxa=%d and number of sites=%d must be positive", *ntax, *nchars);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  sequences=(char**) malloc(*ntax * sizeof(char*));
  num_to_names=(char**) malloc(*ntax * sizeof(char*));
  for(i=0;i<*ntax;i++){
    sequences[i] = (char*) malloc((*nchars+1) * sizeof(char));
    num_to_names[i] = (char*) malloc(MAX_NAME_LEN * sizeof(char));
    fscanf(fp, "%s", num_to_names[i]);
    fscanf(fp, "%s", sequences[i]);
  }

  /*
  for(i=0;i<*ntax;i++)
    printf("%s\t%s\n", num_to_names[i], sequences[i]);
  */

  *sequences_ = sequences;
  *num_to_names_ = num_to_names;
}

