/*
This file contains subroutines used by the recursive-iterative-dcm3 program.
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<constants.h>
#include<phylo_tree.h>
#include<ugraph_aux_vars.h>
#include<dcm3.h>
#include<math.h>
#include<assert.h>
#include<scm_merge.h>
#include<sys/types.h>
#include<unistd.h>
#include<signal.h>
#include<recidcm3.h>
#include<unistd.h>

extern void cleanup(int dummy);
extern int DEBUG;

phylogenetic_tree recursive_dcm3(phylogenetic_tree T, char** sequences, char** num_to_name, hashtable name_to_num, double max_subset_size, char* basecmd, char* basein, char* restricted_gt, char* basetree, int rec_level){
  
  int i, j; phylogenetic_tree U;
  int_set subset_taxa;
  edge c_edge;
  int* subsets; int* subsets_; int* separator;
  int subset_num, separator_size;
  phylogenetic_tree* tree_array;
  phylogenetic_tree recidcm3_tree;

  if(DEBUG) printf("\nEntering level %d with dataset size %d...\n", rec_level, T->num_leaves);
  
  recidcm3_tree = NULL;
  subsets = NULL; subsets_=NULL; separator=NULL;

  subsets_ = (int*) malloc((T->num_leaves+1) * sizeof(int));
  subsets = (int*) malloc(T->num_leaves * sizeof(int));
  separator = (int*) malloc(T->num_leaves * sizeof(int));
  separator_size = 0;

  c_edge = get_centroid_edge(T);
  get_dcm3_subsets(T, c_edge, subsets, subsets_, separator, &subset_num);
  
  tree_array = (phylogenetic_tree*) 
    malloc((subset_num-1) * sizeof(struct phylogenetic_tree_struct));
  
  for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)separator_size++;

  for(i = 0; i < subset_num-1;i++){
    /*if subset size is greater than max_size, then recurse*/
    if(subsets_[i+1] - subsets_[i] + separator_size > max_subset_size){

      /*if subset size is the same size as the tree 
	then we can't go any smaller*/
      if(subsets_[i+1] - subsets_[i] + separator_size >= T->num_leaves){
	/*so we print subsets*/
	subset_taxa = new_int_set(32);
	for(j = subsets_[i]; j < subsets_[i+1]; j++){
	  /*printf("%s\t", num_to_name[subsets[j]]);*/
	  insert_int_set(subset_taxa, subsets[j]);
	}
	for(j = 0; separator[j] != -1 && j < T->num_leaves;j++){
	  /*printf("%s\t", num_to_name[separator[j]]);*/
	  insert_int_set(subset_taxa, separator[j]);
	}
	/*printf("\n");*/

	/*compute subtree*/
	if(DEBUG) printf("Computing subtree of size %d at level %d...\n", size_int_set(subset_taxa), rec_level);
        tree_array[i] = solve_subtree(T, sequences, num_to_name, name_to_num, subset_taxa, basecmd, basein, restricted_gt, basetree);
	if(DEBUG) printf("Computed subtree at level %d\n", rec_level);
	del_int_set(subset_taxa);
      }

      else { /*we recurse*/
	/*convert subset into an integer-set*/

	subset_taxa = new_int_set(32);
	for(j = subsets_[i]; j < subsets_[i+1]; j++)
	  insert_int_set(subset_taxa, subsets[j]);
	/*don't forget the separator*/
	for(j = 0; separator[j] != -1 && j < T->num_leaves;j++)
	  insert_int_set(subset_taxa, separator[j]);
	/*copy tree and restrict it*/
	U = new_phylogenetic_tree(); 
	assign_tree(U, T); 
	restrict_tree(U, subset_taxa);
	/*set subtree sizes*/
	set_subtree_sizes(U);
	/*now call recursive-dcm3*/
	tree_array[i] = recursive_dcm3(U, sequences, num_to_name, name_to_num,  max_subset_size, basecmd, basein, restricted_gt, basetree, rec_level+1);
	del_phylogenetic_tree(U); del_int_set(subset_taxa);
      }

    }
    else {
      /*print subsets*/
      subset_taxa = new_int_set(32);
      for(j = subsets_[i]; j < subsets_[i+1]; j++){
	/*printf("%s\t", num_to_name[subsets[j]]);*/
	insert_int_set(subset_taxa, subsets[j]);	
      }
      for(j = 0; separator[j] != -1 && j < T->num_leaves;j++){
        /*printf("%s\t", num_to_name[separator[j]]);*/
	insert_int_set(subset_taxa, separator[j]);	
      }
      /*printf("\n");*/
      
      /*compute subtree*/
      if(DEBUG) printf("Computing subtree of size %d at level %d...\n", size_int_set(subset_taxa), rec_level);
      tree_array[i] = solve_subtree(T, sequences, num_to_name, name_to_num, subset_taxa, basecmd, basein, restricted_gt, basetree);
      if(DEBUG) printf("Computed subtree at level %d\n", rec_level);
      del_int_set(subset_taxa);
    }
  }

  free(subsets); free(subsets_); free(separator);

  /*merge trees*/
  if(DEBUG) printf("Merging %d trees at level %d...\n", subset_num-1, rec_level);
  recidcm3_tree = scm_merge(tree_array, subset_num - 1);
  if(DEBUG) printf("Done merging\n");

  /*free tree_array*/
  for(i=0;i<subset_num-1;i++) del_phylogenetic_tree(tree_array[i]);
  free(tree_array);

  /*random refine recidcm3 tree*/
  if(DEBUG) printf("Randomly refining tree at level %d...\n",rec_level);
  random_refine(recidcm3_tree);
  if(DEBUG) printf("Refined tree\n");

  if(DEBUG) printf("Returning from level %d\n", rec_level);

  return recidcm3_tree;
}

phylogenetic_tree global_search(char* globalcmd, char** num_to_name, hashtable name_to_num, int_set full_taxa, double* treescore, char* scorefile, char* globalsearchtree){
  
  int result;
  char newick_tree[MAX_NEWICK_STR_LEN];
  phylogenetic_tree T; FILE* fp;
  int_set tree_taxa; node v;

  result = system(globalcmd);
  if(result != 0) {
    printf("Error: could not execute system command \"%s\"\n", globalcmd);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  /*read in tree score*/
  fp = fopen(scorefile, "r");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", scorefile);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%lf", treescore);
  fclose(fp);

  /*read in new tree*/
  fp = fopen(globalsearchtree, "r");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", globalsearchtree);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%s", newick_tree);
  fclose(fp);

  T = new_phylogenetic_tree();
  if(!read_tree(T, num_to_name, newick_tree, name_to_num)) {
  /*if(!read_rooted_tree(T, num_to_name, newick_tree, name_to_num)) {*/
    printf("Error: could not read tree %s\n", newick_tree);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  /*check to make sure taxa of tree read and the full dataset taxa match*/
  tree_taxa = new_int_set(32);
  for(v = T->G->first_node; v; v = v->next)
    if(v->degree==1) {
      assert(((node_aux_vars)v->aux_vars)->taxon_id != -1);
      insert_int_set(tree_taxa, ((node_aux_vars)v->aux_vars)->taxon_id);
    }
  assert(equal_int_set(tree_taxa, full_taxa));
  del_int_set(tree_taxa);

  return T;
}

phylogenetic_tree solve_subtree(phylogenetic_tree T, char** sequences, char** num_to_name, hashtable name_to_num, int_set subset_taxa, char* basecmd, char* basein, char* restricted_gt, char* basetree ){
  
  FILE* fp; int result;
  char newick_tree[MAX_NEWICK_STR_LEN];
  phylogenetic_tree BT, U; int_set tree_taxa;
  node v; int i; int* entries;

  assert(size_int_set(subset_taxa) > 3);
  
  /*find restricted tree and print to file*/
  U = new_phylogenetic_tree(); 
  assign_tree(U, T); 
  restrict_tree(U, subset_taxa);

  /*write tree with branch lengths*/
  for(i=0;i<strlen(newick_tree);i++) newick_tree[i]='\0';
  print_tree(U, num_to_name, newick_tree, 1);
  fp = fopen(restricted_gt, "w");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", restricted_gt);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  fprintf(fp, "%s", newick_tree); fclose(fp);
  del_phylogenetic_tree(U);

  /*print sequences to file*/
  fp = fopen(basein, "w");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", basein);
    cleanup(1);
    exit(EXIT_FAILURE);
  }

  fprintf(fp, "%d %d\n", size_int_set(subset_taxa), strlen(sequences[0]));
  entries=(int*)malloc(size_int_set(subset_taxa) * sizeof(int));
  get_entries_int_set(subset_taxa, entries);
  for(i=0;i<size_int_set(subset_taxa);i++){
    assert(num_to_name[entries[i]]);
    fprintf(fp, "%-30s%s\n", num_to_name[entries[i]], sequences[entries[i]]);
  }
  fclose(fp); free(entries);

  /*printf("base cmd=%s\n", basecmd);*/
  result = system(basecmd);
  if(result != 0) {
    printf("Error: could not execute system command \"%s\"\n", basecmd);
    cleanup(1);
    exit(EXIT_FAILURE);
  }

  /*read in base tree*/
  fp = fopen(basetree, "r");
  if(!fp) {
    printf("Error: could not open filename \"%s\"\n", basetree);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%s", newick_tree);
  fclose(fp);
  
  BT = new_phylogenetic_tree();
  if(!read_tree(BT, num_to_name, newick_tree, name_to_num)){
    printf("Error: could not read tree %s\n", newick_tree);
    cleanup(1);
    exit(EXIT_FAILURE);
  }

  /*check to make sure taxa of tree read and the subset_taxa match*/
  tree_taxa = new_int_set(32);
  for(v = BT->G->first_node; v; v = v->next)
    if(v->degree==1) {
      assert(((node_aux_vars)v->aux_vars)->taxon_id != -1);
      insert_int_set(tree_taxa, ((node_aux_vars)v->aux_vars)->taxon_id);
    }
  if (!equal_int_set(tree_taxa, subset_taxa)){
    printf("Error: taxa in tree sent by base method different from subset taxa\n");
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  del_int_set(tree_taxa);

  return BT;
}
  
void read_chars(FILE* fp, char*** num_to_names_, char*** sequences_,
		int* ntax, int* nchars){
  
  int i; char** sequences; char** num_to_names;

  *ntax=0; *nchars=0;
  fscanf(fp, "%d", ntax);
  fscanf(fp, "%d", nchars);
  if(*ntax<=0 || *nchars<=0){
    printf("Error: number of taxa=%d and number of sites=%d must be positive", *ntax, *nchars);
    cleanup(1);
    exit(EXIT_FAILURE);
  }
  
  sequences=(char**) malloc(*ntax * sizeof(char*));
  num_to_names=(char**) malloc(*ntax * sizeof(char*));
  for(i=0;i<*ntax;i++){
    sequences[i] = (char*) malloc((*nchars+1) * sizeof(char));
    num_to_names[i] = (char*) malloc(MAX_NAME_LEN * sizeof(char));
    fscanf(fp, "%s", num_to_names[i]);
    fscanf(fp, "%s", sequences[i]);
  }

  /*
  for(i=0;i<*ntax;i++)
    printf("%s\t%s\n", num_to_names[i], sequences[i]);
  */

  *sequences_ = sequences;
  *num_to_names_ = num_to_names;
}

