/*
This file merges two set of splits using the Strict Consensus Merger.
Copyright (C) 2004 The University of Texas at Austin

Based upon Daniel Huson's implementation. 

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<split.h>
#include<list.h>
#include<scm.h>
#include<hashtable2.h>
#include<hashtable.h>
#include<stdio.h>
#include<assert.h>
#include<stdlib.h>

static short separating_split(const int r, const split S1,
			      const split S2, const int_set taxa);

/*splits must be empty*/
void scm(int_set taxa1, int_set taxa2, int_set X, 
	 list splits1, list splits2, list splits){
  
  list_node ln1, ln2, ln3;
  split S, new_S, SX, Si, S1, S2;
  int_set A, B, AX, BX, P, Q, U, V;
  double w; int r, found, found_, max, min, top, i, i1, i2, j;
  hashtable2 equiv_classes1, equiv_classes2;
  list induced_splits1, induced_splits2, equiv_class;
  list induced_splits, induced_splits_, templist;

  /*consider all splits in splits1 that don't split X*/
  for(ln1=splits1->head; ln1; ln1=ln1->next){
    S = (split) ln1->data;
    w = S->wgt;
    A = get_A_split(S);
    B = get_B_split(S);
    
    if(subseteq_int_set(X, A)){
      new_S = new_split();
      U = union_int_set(A, taxa2);
      set_split(new_S, U, B, w);
      del_int_set(U);
      push_list(splits, new_S);
    }
    else if(subseteq_int_set(X, B)){
      new_S = new_split();
      U = union_int_set(B, taxa2);
      set_split(new_S, A, U, w);
      del_int_set(U);
      push_list(splits, new_S);
    }
  }
  /*********/

  /*consider all splits in splits2 that don't split X*/
  for(ln1=splits2->head; ln1; ln1=ln1->next){
    S = (split) ln1->data;
    w = S->wgt;
    A = get_A_split(S);
    B = get_B_split(S);
    
    if(subseteq_int_set(X, A)){
      new_S = new_split();
      U = union_int_set(A, taxa1);
      set_split(new_S, U, B, w);
      del_int_set(U);
      push_list(splits, new_S);
    }
    else if(subseteq_int_set(X, B)){
      new_S = new_split();
      U = union_int_set(B, taxa1);
      set_split(new_S, A, U, w);
      del_int_set(U);
      push_list(splits, new_S);
    }
  }
  /*********/

  r=choose_int_set(X);

  /*find equiv classes in splits1*/
  induced_splits1=new_list(); 
  equiv_classes1=new_hashtable2(size_list(splits1));
  for(ln1=splits1->head; ln1; ln1=ln1->next){
    S = (split) ln1->data;
    A = get_A_split(S);
    B = get_B_split(S);
    if(!(subseteq_int_set(X,A) || subseteq_int_set(X,B))){
      AX = intersect_int_set(A, X);
      BX = intersect_int_set(B, X);
      SX = new_split();
      set_split(SX, AX, BX, 1);
      del_int_set(AX); del_int_set(BX);

      /*TO DO: check for duplicate splits using a char* -> void* hashtable
	by storing the sets as strings. right now we need to traverse
	all of induced_splits1 to do that. this takes worse case 
	O(|induced_splits1|) time but can be done in O(1) using a 
	hashtable.*/
      
      found_=0;
      for(ln2=induced_splits1->head;ln2 && !found_;ln2=ln2->next){
	/*check to see if SX already in induced_splits1*/
	A = get_A_split((split)ln2->data);
	B = get_A_split(SX);
	if(equal_int_set(A, B)){
	  found_=1; 
	  equiv_class = (list) lookup_hashtable2(equiv_classes1, (PTR_TYPE) ln2->data);
	  assert(equiv_class);
	  found=0;
	  for(ln3=equiv_class->head; ln3; ln3=ln3->next){
	    Si = (split)ln3->data;
	    if(separating_split(r, S, Si, taxa1)){
	      insert_list(equiv_class, (void*) S, ln3, 0);
	      found=1;
	      break;
	    }
	  }
	  if(!found) append_list(equiv_class, (void*) S);
	  break; /*exit loop to proceed to next split in splits1*/
	}
      }
      if(!found_){
	equiv_class=new_list();
	append_list(equiv_class, (void*) S);
	insert_hashtable2(equiv_classes1, (PTR_TYPE) SX, (void*) equiv_class);
	push_list(induced_splits1, (void*) SX);
      }
      else 
	del_split(SX);
    }
  }
  /***************/

  /*printf("size list induced_splits1 = %d\n", size_list(induced_splits1));*/

  /*find equiv classes in splits2*/
  induced_splits2=new_list(); 
  equiv_classes2=new_hashtable2(size_list(splits2));
  for(ln1=splits2->head; ln1; ln1=ln1->next){
    S = (split) ln1->data;
    A = get_A_split(S);
    B = get_B_split(S);

    if(!(subseteq_int_set(X,A) || subseteq_int_set(X,B))){
      AX = intersect_int_set(A, X);
      BX = intersect_int_set(B, X);
      SX = new_split();
      set_split(SX, AX, BX, 1);
      del_int_set(AX); del_int_set(BX);

      /*TO DO: check for duplicate splits using a char* -> void* hashtable
	by storing the sets as strings. right now we need to traverse
	all of induced_splits2 to do that. this takes worse case 
	O(|induced_splits2|) time but can be done in O(1) using a 
	hashtable.*/
      
      found_=0;
      for(ln2=induced_splits2->head;ln2 && !found_;ln2=ln2->next){
	/*check to see if SX already in induced_splits2*/
	A = get_A_split((split)ln2->data);
	B = get_A_split(SX);
	if(equal_int_set(A, B)){
	  found_=1;
	  equiv_class = (list) lookup_hashtable2(equiv_classes2, (PTR_TYPE) ln2->data);
	  assert(equiv_class);
	  found=0;
	  for(ln3=equiv_class->head; ln3; ln3=ln3->next){
	    Si = (split)ln3->data;
	    if(separating_split(r, S, Si, taxa2)){
	      insert_list(equiv_class, (void*) S, ln3, 0);
	      found=1;
	      break;
	    }
	  }
	  if(!found) append_list(equiv_class, (void*) S);
	  break; /*exit loop to proceed to next split in splits2*/
	}
      }
      if(!found_){
	equiv_class=new_list();
	append_list(equiv_class, (void*) S);
	insert_hashtable2(equiv_classes2, (PTR_TYPE) SX, (void*) equiv_class);
	push_list(induced_splits2, (void*) SX);
      }
      else 
	del_split(SX);
    }
  }
  
  /*printf("size list induced_splits2 = %d\n", size_list(induced_splits2));*/

  /*compute intersection of the induced lists*/
  /*this implementation takes O(|induced_splits1| x |induced_splits2|) 
    time. this runing time depends on the size of the separator*/
  induced_splits=new_list(); induced_splits_ = new_list();
  for(ln1=induced_splits1->head; ln1; ln1=ln1->next){
    A = get_A_split((split)ln1->data);
    for(ln2=induced_splits2->head; ln2; ln2=ln2->next){
      B = get_A_split((split)ln2->data);
      if(equal_int_set(A, B)){
	push_list(induced_splits, (void*) ln1);
	push_list(induced_splits_, (void*) ln2);
	assert(lookup_hashtable2(equiv_classes1, (PTR_TYPE) ln1->data));
	assert(lookup_hashtable2(equiv_classes2, (PTR_TYPE) ln2->data));
      }
    }
  }

  /*printf("size of induced splits = %d\n", size_list(induced_splits));
  printf("size of induced splits_ = %d\n", size_list(induced_splits_));*/

  ln2=induced_splits_->head;
  for(ln1=induced_splits->head; ln1; ln1=ln1->next){
    list equiv_class1=(list)lookup_hashtable2(equiv_classes1, (PTR_TYPE) ((list_node)ln1->data)->data);
    list equiv_class2=(list)lookup_hashtable2(equiv_classes2, (PTR_TYPE) ((list_node)ln2->data)->data);
    assert(equiv_class1);
    assert(equiv_class2);

    /*TO DO: convert both lists to arrays so that accessing ith 
      entry is faster*/

    if(size_list(equiv_class1)>size_list(equiv_class2)){
      max=size_list(equiv_class1); min = size_list(equiv_class2);}
    else {max=size_list(equiv_class2);min=size_list(equiv_class1);}
    
    /*if(min>1 && _SCM_MERGE_METHOD_==0) { ln2=ln2->next; continue; }*/
    if(min>1){ top = 2; } else { top = max; }

    for(i=0;i<top;i++){
      if(top==2) {
        if(i!=0){
         assert(size_list(equiv_class1) && size_list(equiv_class2));
         i1 = size_list(equiv_class1) - 1;
         i2 = size_list(equiv_class2) - 1;
        }
        else { i1 = 0; i2 = 0; }
      }
      else{
	i1=(int)((double)size_list(equiv_class1)*((double)i/(double)top));
        i2=(int)((double)size_list(equiv_class2)*((double)i/(double)top));
      }
      
      ln3=equiv_class1->head;
      for(j=0;j<size_list(equiv_class1);j++){
	if(j==i1) { S1=(split)ln3->data; break; }
	ln3=ln3->next;
      }
      ln3=equiv_class2->head;
      for(j=0;j<size_list(equiv_class2);j++){
	if(j==i2) { S2=(split)ln3->data; break; }
	ln3=ln3->next;
      }
    
      if(S1->wgt < S2->wgt) 
	w = S1->wgt;
      else 
	w = S2->wgt;

      A=get_A_split(S1);
      B=get_B_split(S1);

      P=get_A_split(S2);
      Q=get_B_split(S2);

      if(member_int_set(A, r) == member_int_set(P, r)){
	S = new_split();
	U = union_int_set(A, P); V = union_int_set(B, Q);
	set_split(S, U, V, w);
	del_int_set(U); del_int_set(V);
	push_list(splits, (void*) S);
      }
      else {
	S = new_split();
	U = union_int_set(A, Q); V = union_int_set(B, P);
	set_split(S, U, V, w);
	del_int_set(U); del_int_set(V);
	push_list(splits, (void*) S);
      }
    }
    ln2=ln2->next;
  }

  /*free induced_list1, induced_list2, induced_list, induce_list_,
    and all hashtables*/
  for(ln1=induced_splits1->head; ln1; ln1=ln1->next){
    del_split((split)ln1->data); ln1->data=NULL;
  }
  del_list(induced_splits1);
  induced_splits1=NULL;

  for(ln2=induced_splits2->head; ln2; ln2=ln2->next){
    del_split((split)ln2->data); ln2->data=NULL;
  }
  del_list(induced_splits2);
  induced_splits2=NULL;

  for(ln1=induced_splits->head; ln1; ln1=ln1->next){
    ln1->data=NULL;
  }
  for(ln1=induced_splits_->head; ln1; ln1=ln1->next){
    ln1->data=NULL;
  }

  del_list(induced_splits); del_list(induced_splits_);
  
  for(i=0;i<equiv_classes1->size_;i++)
    if(equiv_classes1->table_[i] && equiv_classes1->table_[i]->value){
      templist = (list) equiv_classes1->table_[i]->value;
      for(ln1=templist->head; ln1; ln1=ln1->next) ln1->data=NULL;
      del_list(templist);
      equiv_classes1->table_[i]->value = NULL;
    }
  del_hashtable2(equiv_classes1);
  equiv_classes1=NULL;

  for(i=0;i<equiv_classes2->size_;i++)
    if(equiv_classes2->table_[i] && equiv_classes2->table_[i]->value){
      templist = (list) equiv_classes2->table_[i]->value;
      for(ln1=templist->head; ln1; ln1=ln1->next) ln1->data=NULL;
      del_list(templist);
      equiv_classes2->table_[i]->value = NULL;
    }
  del_hashtable2(equiv_classes2);
  equiv_classes2=NULL;

}

static short separating_split(const int r, const split S1,
			      const split S2, const int_set taxa){
  int_set A, B, P, Q;
  int ret_value;
  
  A = get_A_split(S1); B = get_B_split(S1);

  P = get_A_split(S2); Q = get_B_split(S2);
  
  if(member_int_set(A, r))
    if(member_int_set(P, r))
      ret_value = subseteq_int_set(A, P);
    else
      ret_value = subseteq_int_set(A, Q);
  else
    if(member_int_set(P, r))
      ret_value = subseteq_int_set(B, P);
    else
      ret_value = subseteq_int_set(B, Q);
  
  /*del_int_set(B); del_int_set(Q);
  B=NULL; Q=NULL;*/

  return ret_value;
}
	
