/*
This file merges a set of trees by calling SCM on two at a time.
Copyright (C) 2004 The University of Texas at Austin

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 59
Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Email: usman@cs.njit.edu
Mail: Usman Roshan, Department of Computer Sciences, NJIT-CCS, Computer
Science Department, GITC 4400, University Heights, Newark, NJ 07102
*/
#include<string.h>
#include<stdio.h>
#include<stdlib.h>
#include<tree2splits.h>
#include<splits2tree.h>
#include<constants.h>
#include<phylo_tree.h>
#include<assert.h>
#include<list.h>
#include<split.h>
#include<scm.h>
#include<ugraph_aux_vars.h>
#include<time.h>
#include<scm_merge.h>

static void add_trivial_splits(list scm_splits, int_set taxa);

phylogenetic_tree scm_merge(phylogenetic_tree* tree_array, int num_trees){
  
  int i; 
  list* splits_array;
  int_set* taxa_array;
  int t2s_time, s2t_time, scm_time;
  time_t begin; time_t end; /* for profiling */
  node v;
  list_node ln;
  list scm_splits;
  phylogenetic_tree scm_tree;
  int_set X, taxa, tempset;
  
  assert(num_trees > 1);
  splits_array = (list*) malloc(num_trees * sizeof(list));
  taxa_array = (int_set*) malloc(num_trees * sizeof(int_set));
  
  /*computing set of taxa in each tree*/
  for(i=0;i<num_trees;i++){
    taxa_array[i] = new_int_set(32);
    for(v = tree_array[i]->G->first_node; v; v=v->next){
      if(v->degree == 1){
	insert_int_set(taxa_array[i], ((node_aux_vars)v->aux_vars)->taxon_id);
      }
    }
  }
  /****************/
  
  /*printf("doing tree2splits\n");*/
  begin=time(NULL);
  for(i=0;i<num_trees;i++){
    splits_array[i] = new_list();
    tree2splits(tree_array[i], splits_array[i]);
  }
  end=time(NULL); t2s_time+=difftime(end,begin);
  /*printf("ending tree2splits\n");*/

  begin=time(NULL);
  taxa=new_int_set(32); assign_int_set(taxa, taxa_array[0]);
  scm_splits = new_list(); 
  for(i=1; i<num_trees; i++){
    X = intersect_int_set(taxa, taxa_array[i]);
    /*merge is returned in scm_splits, X is intersection,
      taxa is set of taxa of splits_array[i-1] and taxa_array[i]
      is set of taxa of splits_array[i]*/
    scm(taxa, taxa_array[i], X, splits_array[i-1], splits_array[i],
	scm_splits);

    del_int_set(X); X=NULL; /*don't need X anymore*/

    /*update taxa*/
    tempset = union_int_set(taxa, taxa_array[i]);
    assign_int_set(taxa, tempset); 
    del_int_set(tempset); tempset=NULL;

    /*delete splits_array[i] and splits_array[i-1] 
      --- don't need both of them anymore*/
    for(ln=splits_array[i]->head;ln;ln=ln->next){
      if(ln->data) del_split((split)ln->data); ln->data=NULL;
    }
    del_list(splits_array[i]); splits_array[i] = NULL;
    if(splits_array[i-1] != NULL){
      for(ln=splits_array[i-1]->head;ln;ln=ln->next){
	if(ln->data) del_split((split)ln->data); ln->data=NULL;
      }
      del_list(splits_array[i-1]); splits_array[i-1] = NULL;
    }

    /*updated splits_array only if not reached the end*/
    if(i != num_trees-1){
      /*set splits_array[i] to newly computed scm_splits*/
      splits_array[i] = scm_splits;
      /*reset scm_splits variable*/
      scm_splits = new_list();
    }
  }
  /*del_int_set(X);*/
  end=time(NULL); scm_time+=difftime(end,begin);

  /*adding trivial splits*/
  add_trivial_splits(scm_splits, taxa);
  
  /*deleting arrays*/
  /*tree_array we should delete in the calling program*/
  for(i=0;i<num_trees;i++) {
    del_int_set(taxa_array[i]); taxa_array[i]=NULL;
    /*del_phylogenetic_tree(tree_array[i]);*/
    if(splits_array[i] != NULL){
      for(ln=splits_array[i]->head; ln; ln=ln->next){
	del_split((split)ln->data); ln->data=NULL;
      }
      del_list(splits_array[i]);
    }
  }
  free(taxa_array);
  free(splits_array);
  /****************/

  scm_tree = new_phylogenetic_tree();

  begin=time(NULL);
  /*printf("calling splits2tree\n");*/
  splits2tree(scm_splits, taxa, scm_tree);
  /*printf("done from splits2tree\n");*/
  end=time(NULL); s2t_time+=difftime(end,begin);
  
  /*delete scm_splits*/
  for(ln=scm_splits->head; ln; ln=ln->next){
    assert((split)ln->data); del_split((split)ln->data); 
    ln->data=NULL;
  }
  del_list(scm_splits);
  del_int_set(taxa);
 
  /*printf("scm tree has %d nodes and %d edges\n", scm_tree->G->n_nodes,
	 scm_tree->G->n_edges);*/
  return scm_tree;
}

static void add_trivial_splits(list scm_splits, int_set taxa){
  
  int_set A, B, taxa2; list_node ln; int i, t; 
  split S; int* intset_entries;

  taxa2=new_int_set(32);
  assign_int_set(taxa2, taxa);
  for(ln=scm_splits->head; ln; ln=ln->next){
    A = get_A_split((split)ln->data);
    if(size_int_set(A)==1)
      remove_int_set(taxa2, choose_int_set(A));
    else if (size_int_set(A)==size_int_set(taxa)-1){
      B=get_B_split((split)ln->data);
      remove_int_set(taxa2, choose_int_set(B));
    }
  }
  intset_entries = (int*) malloc(size_int_set(taxa2) * sizeof(int));
  get_entries_int_set(taxa2, intset_entries);
  for(i=0;i<size_int_set(taxa2);i++){
    t = intset_entries[i];
    S = new_split();
    B = new_int_set(32);
    assign_int_set(B, taxa);
    A = new_int_set(32);
    insert_int_set(A, t);
    remove_int_set(B, t);
    set_split(S, A, B, 1);
    del_int_set(A); del_int_set(B);
    append_list(scm_splits, S);
  }
  free(intset_entries);
  del_int_set(taxa2);
}
